// SpadeLover Spades -- A Spades Game
// Copyright (C) 1998 Britt Yenne
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// To contact the author, visit:  http://www.jump.net/~syenne/pilot/

#if defined(_AIX) || defined(linux)
# include <stdio.h>
#else
# pragma pack(2)
# include <Common.h>
# include <System/SysAll.h>
#endif

#include "header.h"

#define AUTOPLAY

#ifdef DEBUG
static char *
CardName(int c)
{
  static char buf[4];

  buf[0] = "23456789TJQKA"[Rank(c)];
  buf[1] = "shcd"[Suit(c)];
  buf[2] = '\0';
  return buf;
}
#endif

/*
 * Reset all game preferences to default values.  The given game
 * parameter is the game number to reset, or -1 for all games.
 */
void
SetDefaultPrefs(int game)
{
  int g = 0;
  struct _gpref normal;

  /*
   * First init our idea of a "normal" game of partner spades.
   */
  bzero(&normal, sizeof(normal));
  normal.gprefWin = 500;
  normal.gprefLose = -300;
  normal.gprefMaxInd = 13;
  normal.gprefMaxTeam = 13;
  normal.gprefPartner = 1;
  normal.gprefSpadeBreak = 1;

  /*
   * Game 0: Partner Spades
   */
  if (game < 0 || game == g) {
    S.gpref[g] = normal;
    S.gpref[g].gprefBlindNils = 1;
  }
  g++;

  /*
   * Game 1: Solo Spades
   */
  if (game < 0 || game == g) {
    bzero(&S.gpref[g], sizeof(S.gpref[g]));
    S.gpref[g].gprefWin = 300;
    S.gpref[g].gprefLose = 0;		// can't lose on score
    S.gpref[g].gprefMaxInd = 13;
    S.gpref[g].gprefSpadeBreak = 1;
    S.gpref[g].gprefClubTrick = 1;
    S.gpref[g].gprefMustBeat = 1;
    S.gpref[g].gprefBlindNils = 1;
  }
  g++;

  /*
   * Game 2: Suicide Spades
   */
  if (game < 0 || game == g) {
    S.gpref[g] = normal;		// start with partner spades defaults
    S.gpref[g].gprefWin = 300;		// play to 300
    S.gpref[g].gprefMinTeam = 4;	// team must bid at least 4
    S.gpref[g].gprefOneNil = 1;		// one teammate must go nil
  }
  g++;

  /*
   * Game 3: Mirrors Spades
   */
  if (game < 0 || game == g) {
    S.gpref[g] = normal;		// start with partner spades defaults
    S.gpref[g].gprefBidSpades = 1;	// must bid spade count
  }
  g++;

  /*
   * Remaining games are custom and default to partner spades
   */
  for (; g < nGames; g++)
    if (game < 0 || game == g)
      S.gpref[g] = normal;
}

static int
SetState()
{
  if (S.turn == 0)
    return (S.p[0].bid < 0) ? StUserBid : StUserPlay;
  return 0;
}

void
SortCards(s8bit *card, int count)
{
  int i, j, suit, step, n, p;
  s8bit tmp;

  if (S.prefRevSuits) {
    suit = 3;
    step = -1;
  }
  else {
    suit = 0;
    step = 1;
  }

  for (; suit >= 0 && suit < 4; suit += step)
  {
    for (i = n = 0; i < count; i++)
      if (Suit(card[i]) == suit)
      {
	tmp = card[i];
	card[i] = card[n];
	card[n++] = tmp;
      }
    for (i = 0; i < n; i++)
    {
      for (p = i, j = i + 1; j < n; j++)
	if ((S.prefRevRanks && card[j] < card[p])
	 || (!(S.prefRevRanks) && card[j] > card[p]))
	  p = j;
      tmp = card[p];
      card[p] = card[i];
      card[i] = tmp;
    }
    card += n;
    count -= n;
  }
}

void
NewHand()
{
  int i, j, p, c;
  s8bit card[52];
  struct score score;

  /*
   * Score previous hand.
   */
  if (S.state == StEndHand) {
    for (p = 0; p < 4; p++) {
      Score(p, &score);
      S.p[p].score += score.totalScore;
      S.p[p].bags = (S.p[p].bags + score.bags) % 10;

      if (S.p[p].score >= S.prefWin)
	S.state = StEndGame;
      else if (S.prefWin > S.prefLose) {
	if (S.prefLose != 0 && S.p[p].score <= S.prefLose)
	  S.state = StEndGame;
      }
      else if (S.p[p].score >= S.prefLose)
	S.state = StEndGame;
    }
  }
  if (S.state == StEndGame)
    return;

  /*
   * Next dealer, please.
   */
  if (++S.dealer >= 4)
    S.dealer = 0;
  if ((S.turn = S.dealer + 1) >= 4)
    S.turn = 0;

  /*
   * Shuffle and deal.
   */
  for (i = 0; i < 52; i++)
    card[i] = i;
  for (i = 0, p = 0; p < 4; p++)
    for (c = 0; c < 13; c++, i++) {
      j = i + (SysRandom(0) % (52 - i));
      S.p[p].card[c] = card[j];
      card[j] = card[i];
    }
  SortCards(S.p[0].card, 13);
#ifdef DEBUG
  SortCards(S.p[1].card, 13);
  SortCards(S.p[2].card, 13);
  SortCards(S.p[3].card, 13);
#endif

  /*
   * Reset state variables.
   */
  for (p = 0; p < 4; p++) {
    S.p[p].bid = -1;	// ie: hasn't bid yet
    S.p[p].blind = 0;
    S.p[p].taken = 0;
    S.p[p].play = -1;	// ie: hasn't played yet
    S.p[p].last = -1;	// ie: hasn't played yet
    for (i = 0; i < 4; i++) {
      S.p[p].suit[i] = 0x1fff;	// he could have any card
      S.p[p].pmask[i] = 0;
    }
    S.played[p] = 0;	// 4 players, 4 suits, right?
  }
  S.state = SetState();
}

void
NewGame()
{
  int p;

  S.dealer = 3;		// ie. player will be next (first) dealer
  for (p = 0; p < 4; p++) {
    S.p[p].score = 0;
    S.p[p].bags = 0;
  }
  S.state = 0;
  NewHand();
}

void
NewTrick()
{
  int p;

  for (p = 0; p < 4; p++)
    S.p[p].play = -1;
  S.state = SetState();
}

static int
HasCard(int p, int c)
{
  int i;

  for (i = 0; i < 13; i++)
    if (S.p[p].card[i] == c)
      return 1;
  return 0;
}

void
MakeBid(int p, int bid)
{
  int pard, tot, orig, i, blind;

  /*
   * Handle special case for blind nils.
   */
  if (bid == bidBlindNil) {
    bid = 0;
    blind = 1;
  }
  else
    blind = 0;
  orig = bid;

#if defined(DEBUG) && 0
  printf("P%d bids %s%d\n", p, (blind ? "blind " : ""), bid);
#endif

  /*
   * Tweak team bids per preferences.
   */
  if (S.prefBidSpades) {
    bid = 0;
    for (i = 0; i < 13; i++)
      if (Suit(S.p[p].card[i]) == 0)
	bid++;
  }
  else if (S.prefPartner) {
    pard = (p + 2) % 4;
    if (bid > S.prefMaxTeam)
      bid = S.prefMaxTeam;
    if (bid > 0 && bid < S.prefMinTeam && S.prefOneNil)
      bid = S.prefMinTeam;

    if (S.p[pard].bid >= 0) {
      if (S.prefOneNil && S.p[pard].bid > 0 && bid > 0) {
	S.p[pard].bid = 0;
	BidChanged(pard, p);
      }

      tot = bid + S.p[pard].bid;
      if (tot < S.prefMinTeam) {
	tot = (int)S.prefMinTeam - tot;
	if (bid > 0) {
	  bid += tot;
	  if (bid <= S.prefMaxInd)
	    tot = 0;
	  else {
	    tot = bid - S.prefMaxInd;
	    bid = S.prefMaxInd;
	  }
	}
	if (tot > 0 && S.p[pard].bid < S.prefMaxInd) {
	  S.p[pard].bid += tot;
	  if (S.p[pard].bid <= S.prefMaxInd)
	    tot = 0;
	  else {
	    tot = S.p[pard].bid - S.prefMaxInd;
	    S.p[pard].bid = S.prefMaxInd;
	  }
	  BidChanged(pard, p);
	}
	if (tot > 0) {
	  bid += tot;
	  if (bid > S.prefMaxInd)
	    bid = S.prefMaxInd;
	}
      }
      else if (tot > S.prefMaxTeam) {
	tot -= (int)S.prefMaxTeam;
	if (bid > 0) {
	  if (tot >= bid) {
	    tot -= bid - 1;
	    bid = 1;
	  }
	  else {
	    bid -= tot;
	    tot = 0;
	  }
	}
	if (bid < S.prefMinInd) {
	  tot += S.prefMinInd - bid;
	  bid = S.prefMinInd;
	}
	if (tot > 0 && S.p[pard].bid > S.prefMinInd) {
	  S.p[pard].bid -= tot;
	  if (S.p[pard].bid < S.prefMinInd)
	    S.p[pard].bid = S.prefMinInd;
	  BidChanged(pard, p);
	}
      }
    }
  }

  /*
   * If he bids nil, figure he doesn't have the ace'o'spades.  This
   * doesn't hurt because if he does then he's set anyway.  This will
   * help his pard realize he can cover with the king.
   */
  if (bid == 0)
    S.p[p].suit[0] &= ~(1<<12);

  /*
   * Register the bid.
   */
  S.p[p].bid = bid;
  if (bid == 0 && blind)
    S.p[p].blind = 1;
  if (orig != bid)
    BidChanged(p, p);
  if (++(S.turn) >= 4)
    S.turn = 0;

  /*
   * If this is the end of the bidding, then the person with the two
   * of clubs should go first when prefs are set that way.
   */
  if (S.prefClubTrick && S.p[S.turn].bid >= 0) {
    for (p = 0; p < 4; p++) {
      for (i = 0; i < 13; i++)
	if (S.p[p].card[i] == 26) { // two of clubs
	  S.turn = p;
	  break;
	}
      if (i < 13)
	break;
    }
  }
  S.state = SetState();

  /*
   * Handle blind nil exchanges.
   */
  if (S.prefPartner && S.prefBnilExch > 0 && S.p[S.turn].bid >= 0) {
    if ((S.p[1].bid == 0 && S.p[1].blind) || (S.p[3].bid == 0 && S.p[3].blind))
      BnilExchange(1, NULL);
    if ((S.p[0].bid == 0 && S.p[0].blind) || (S.p[2].bid == 0 && S.p[2].blind))
      S.state = StBnilExch;
  }
}

static void
BnilExchSelect(int p, s8bit *card)
{
  int e, s, i, ee, cover, pct, score, hiScore;
  s8bit worst;

  /*
   * If our bid is 0, find the cards most likely to set us.  Otherwise,
   * find the cards most likely to set our pard's nil.
   */
  for (e = 0; e < S.prefBnilExch; e++) {
    /*
     * We're either going to exchange the highest or lowest card in one
     * of our suits depending on whether we're the nil bidder.  Score each
     * highest (or lowest) and determine which needs to go.
     */
    hiScore = -1;
    for (s = 0; s < 4; s++) {
      worst = -1;
      cover = 0;
      for (i = 0; i < 13 && S.p[p].card[i] >= 0; i++) {
	if (Suit(S.p[p].card[i]) != s)
	  continue;
	for (ee = 0; ee < e; ee++)
	  if (S.p[p].card[i] == card[ee]) {
	    cover++;
	    break;
	  }
	if (ee < e)
	  continue;	// already exchanged this one
	if (s != 0)
	  if ((S.p[p].bid == 0 && Rank(S.p[p].card[i]) <= 6)
	   || (S.p[p].bid != 0 && Rank(S.p[p].card[i]) >= 6))
	    cover++;
	if (worst < 0
	 || (S.p[p].bid == 0 && S.p[p].card[i] > worst)
	 || (S.p[p].bid != 0 && S.p[p].card[i] < worst))
	  worst = S.p[p].card[i];
      }
      if (worst >= 0) {
	score = Rank(worst) * 100 / 12;
	if (S.p[p].bid != 0)
	  score = 100 - score;
	if (s == 0 && S.p[p].bid != 0)
	  pct = 0;		// should never give up trump
	else
	  switch (cover) {
	    case 0:	pct = 100;	break;
	    case 1:	pct = 95;	break;
	    case 2:	pct = 75;	break;
	    case 3:	pct = 50;	break;
	    default:	pct = 25;	break;
	  }
	score = score * pct;
	if (score > hiScore) {
	  card[e] = worst;
	  hiScore = score;
	}
      }
    }
  }
}

void
BnilExchange(int p, s8bit *card)
{
  int e, i, pard, index, pindex, r, s, c;
  s8bit cardBuf[maxExch], pcard[maxExch], tmp;

  /*
   * Find the cards to exchange.
   */
  pard = (p + 2) % 4;
  if (card == NULL) {
    card = cardBuf;
    BnilExchSelect(p, card);
  }
  BnilExchSelect(pard, pcard);

  /*
   * Exchange them.
   */
  for (e = 0; e < S.prefBnilExch; e++) {
    pindex = index = -1;
    for (i = 0; i < 13 && S.p[p].card[i] >= 0; i++)
      if (S.p[p].card[i] == card[e]) {
	index = i;
	break;
      }
    for (i = 0; i < 13 && S.p[pard].card[i] >= 0; i++)
      if (S.p[pard].card[i] == pcard[e]) {
	pindex = i;
	break;
      }
    if (index >= 0 && pindex >= 0) {
      /*
       * Since highest and lowest are generally exchanged, the pards
       * now know something about each others hands.  This could be
       * reflected in the suit map, but then our opponents would know
       * about it too.  The solution: a partner-specific mask.
       */
      if (S.prefExpert) {
	/*
	 * The nil bidder has nothing above the exchanged card in
	 * its suit.
	 */
	if (S.p[p].bid == 0) {
	  i = p;
	  c = S.p[p].card[index];
	}
	else {
	  i = pard;
	  c = S.p[pard].card[pindex];
	}
	s = Suit(c);
	for (r = Rank(c) + 1; r < 13; r++)
	  S.p[i].pmask[s] |= (1<<r);

	/*
	 * The non-nil bidder has nothing below the exchanged card in
	 * its suit.
	 */
	if (S.p[p].bid > 0) {
	  i = p;
	  c = S.p[p].card[index];
	}
	else {
	  i = pard;
	  c = S.p[pard].card[pindex];
	}
	s = Suit(c);
	for (r = Rank(c) - 1; r >= 0; r--)
	  S.p[i].pmask[s] |= (1<<r);
      }

      /*
       * Swap the cards.
       */
      tmp = S.p[p].card[index];
      card[e] = S.p[p].card[index] = S.p[pard].card[pindex];
      S.p[pard].card[pindex] = tmp;
    }
  }
  CardsExchanged(p);
  S.state = SetState();
}

static void
Bid(int p)
{
  int points, i, c, s, gap, p1, p2, pard, prob, bid = -1, nils;
  int suitCount[4], suitAvg[4], count, lowClub, myScore, hiScore;
  int voids = 0, singletons = 0, doubletons = 0;
  int nilrisk = 99;

  /*
   * Should we bid blind nil?
   */
  if (CanBidBlindNil(p)) {
    if (S.prefPartner) {
      myScore = S.p[p%2].score;
      hiScore = S.p[(p+1)%2].score;
    }
    else {
      hiScore = myScore = S.p[p].score;
      for (i = 0; i < 4; i++)
	if (S.p[i].score > hiScore)
	  hiScore = S.p[i].score;
    }
    pard = (S.prefPartner) ? S.p[(p+2)%4].bid : -1;

    if (pard != 0) {
      nils = 0;
      for (i = 0; i < 4; i++) {
	if (S.p[i].bid == 0)
	  nils++;
	else if (S.p[i].bid > 0)
	  count += S.p[i].bid;
	else if (i != p)
	  count += 2;
      }
      if (S.prefPartner && nils == 1)
	count += 3;
      if ((count >= 12 && myScore + 100 <= hiScore)
       || (S.prefBnilExch >= 1 && (count >= 10 || pard >= 5))
       || (S.prefBnilExch >= 1 && pard >= 4 && myScore + 100 <= hiScore)
       || (S.prefBnilExch >= 2 && (count >= 9 || pard >= 4))
       || (S.prefBnilExch >= 2 && pard >= 3 && myScore + 100 <= hiScore)) {
	MakeBid(p, bidBlindNil);
	return;
      }
    }
  }

  /*
   * First count our cards in each suit, and determine the average
   * cards the other players each have in each suit.  Also count our
   * voids, singletons, and doubletons.
   */
  for (s = 0; s < 4; s++)
    suitCount[s] = 0;
  for (i = 0; i < 13; i++)
    suitCount[Suit(S.p[p].card[i])]++;
  for (s = 0; s < 4; s++) {
    if (s > 0)
      switch (suitCount[s]) {
	case 0:	voids++;	break;	// a void
	case 1:	singletons++;	break;	// a singleton
	case 2:	doubletons++;	break;	// a doubleton
      }
    suitAvg[s] = (13 - suitCount[s]) / 3;
  }

  /*
   * Find the rank of our lowest club.
   */
  lowClub = -1;
  for (c = 26; c < 39; c++)	// 2 of clubs thru ace of clubs
    if (HasCard(p, c)) {
      lowClub = Rank(c);
      break;
    }

  // --------------- TRUMPS -----------------
  /*
   * Trumps serve two purposes -- first, any trump sequence down from
   * some high trump may be automatic tricks.  The remaining trumps may
   * either be useful as high trumps or for trumping other tricks if we
   * have some voids, singletons, or doubletons.
   *
   * First, figure out the brute force points possible from trumps:
   */
  p1 = p2 = 0;
  for (c = 12, gap = 0; c >= 0; c--) {
    if (HasCard(p, c)) {
      if (gap > 0)
	gap--;
      else {
	p1 += 100;
      }
    }
    else
      gap++;
  }

  /*
   * If our trumps are losable (ie. p1 == 0) and we don't have many
   * of them, then consider whether we have low enough cards in
   * each of the other suits to risk bidding nil.
   */
  if (p1 == 0 && (S.prefOneNil || S.prefMinInd == 0)) {
    nilrisk = 0;
    if (suitCount[0] > 3)
      nilrisk += suitCount[0] - 2;
    for (s = 0; s < 4; s++) {
      if (suitCount[s] == 0) {
	if (s != 0)
	  nilrisk -= 2;
	continue;
      }
      c = (s * 13);		// the 2 of that suit
      for (i = count = 0; i < 7; i++, c++) {
	if (HasCard(p, c))
	  count++;		// count low cards
      }
      if (count >= 3)		// suit is covered
	continue;
      nilrisk += suitCount[s] - count;	// risky high cards
      if (count == 0)
	nilrisk++;			// no low cards
      if (s == 0) {
	if (HasCard(p, 11))	// king of spades
	  nilrisk += 2;
	if (HasCard(p, 10))	// queen of spades
	  nilrisk += 1;
      }
      else {
	if (HasCard(p, (s * 13) + 11) && suitCount[s] <= 3)
	  nilrisk += ((suitCount[s] <= 2) ? 2 : 1);	// king is riskier
	if (HasCard(p, (s * 13) + 12) && suitCount[s] <= 3)
	  nilrisk += ((suitCount[s] <= 2) ? 3 : 2);	// ace is riskiest
      }
    }

    /*
     * Add some risk for a high club.
     */
    if (S.prefClubTrick && lowClub >= 6)
      nilrisk += (lowClub >= 8) ? 2 : 1;

    /*
     * Add some risk for voids/singletons/doubletons if must beat highest.
     */
    if (S.prefMustBeat && suitCount[0] > 0) {
      nilrisk += 3 * voids;
      nilrisk += 2 * singletons;
      nilrisk += 1 * doubletons;
    }
  }

  /*
   * That done, frob p1 to count probable trump tricks.
   */
  if (p1 == 0) {
    if (HasCard(p, 11) && suitCount[0] >= 2)	// King with 1+ covers
      p1 += ((suitCount[0] >= 3) ? 100 : 60);
    if (HasCard(p, 10) && suitCount[0] >= 3)	// Queen with 2+ covers
      p1 += ((suitCount[0] >= 4) ? 100 : 30);
  }

  /*
   * Now figure if we use our trumps on other suits:
   */
  count = suitCount[0];
  for (c = 12; c >= 0 && HasCard(p, c); c--) {
    p2 += 100;
    count--;
  }
  for (; voids > 0 && count > 0; voids--) {
    if (count >= 3) {
      p2 += 225;
      count -= 2;	// yes, 2
    }
    else if (count >= 2) {
      p2 += 190;
      count -= 2;
    }
    else {
      p2 += 90;
      count -= 1;
    }
  }
  for (; singletons > 0 && count > 0; singletons--) {
    if (count >= 2) {
      p2 += 175;
      count -= 2;
    }
    else {
      p2 += 90;
      count -= 1;
    }
  }
  for (; doubletons > 0 && count > 0; doubletons--) {
    p2 += 75;
    count -= 1;
  }
  if (count > suitAvg[0]) {
    /*
     * Length DOES matter.  :-)
     */
    p2 += (count - suitAvg[0] - 1) * 80;
  }
  else if (count == 0)
    p2 -= 30;		// be wary of exhausting all trump

  /*
   * Okay, choose the better of the two scores.
   */
  points = (p1 > p2) ? p1 : p2;

  // --------------- OTHER SUITS -----------------
  /*
   * We bid other suits using a confidence system.  We combine two
   * factors:  the confidence we can make some high card good, and
   * the probability of that trick not getting trumped/finessed.
   */
  for (s = 1; s < 4; s++) {
    c = (s * 13) + 12;		// the ace of that suit
    if ((count = suitAvg[s]) > suitCount[s])
      count = suitCount[s];	// the number of probable tricks w/o trump
    prob = 100;			// probability of not getting finessed
    for (; count > 0; count--, c--) {
      if (HasCard(p, c)) {
	points += prob;
	prob = prob * 9 / 10;
      }
      else
	prob = prob * 8 / 10;
    }
  }

  /*
   * Consider opponents' and pard's bids.
   */
  if ((i = p + 1) >= 4)
    i -= 4;
  p1 = S.p[i].bid;
  if ((i = p + 2) >= 4)
    i -= 4;
  pard = (S.prefPartner) ? S.p[i].bid : -1;
  if ((i = p + 3) >= 4)
    i -= 4;
  p2 = S.p[i].bid;

#if defined(DEBUG) && 0
  printf("BID%d: p1=%d pard=%d p2=%d nilrisk=%d points=%d\n",
    p, p1, pard, p2, nilrisk, points);
#endif

  // --------------- NIL BIDS -----------------
  /*
   * Consider bidding nil.
   */
  if (S.prefPartner && (p1 == 0 || p2 == 0 || S.prefOneNil))
    nilrisk -= 2;
  if (bid < 0 && S.prefOneNil && pard > 0 && nilrisk <= 10)
    bid = 0;
  if (bid < 0 && pard != 0 && (nilrisk <= 0
   || (nilrisk <= 1 && points < 250)
   || (nilrisk <= 2 && points < 150)
   || (pard >= 3 + nilrisk)
   || (S.prefOneNil && nilrisk < 6 && points < 450)))
    bid = 0;

  // --------------- NORMAL BIDS -----------------
  /*
   * Add points for a high club.
   */
  if (S.prefClubTrick && lowClub >= 6 && suitCount[2] <= 6 && lowClub < 13 - suitCount[2])
    points += (lowClub >= 8) ? 100 : 75;

  /*
   * Okay, divide points by 100 and that's our bid.
   */
  if (bid < 0) {
    if (S.prefOneNil)
      prob = 130;
    else if (pard == 0)
      prob = 70;
    else if (pard > 4)
      prob = 50;
    else
      prob = 30;
    bid = (points + prob) / 100;

    /*
     * Consider ramifications of others' bids if everyone else has
     * bid.
     */
    count = bid;
    for (i = 0; i < 4; i++) {
      if (i == p)
	continue;
      if (S.p[i].bid < 0)
	break;		// hasn't bid yet -- break
      count += S.p[i].bid;
    }
    if (i >= 4) {		// everyone has bid
      /*
       * If the count is real high, reconsider potential tricks.
       * If the count is high and they bid nil, consider bidding less.
       */
      if (count >= 11) {
	if (count >= 13)
	  bid = points / 100;
	else if (p1 == 0 || p2 == 0)
	  bid = points / 100;
      }
    }
  }

  /*
   * Declare a misdeal?
   */
  if (ValidMisdeal(p) && bid > 0 && bid < 3 && (bid + pard) < 5) {
    NewHand();
    DeclareMisdeal(p);
    return;
  }

  /*
   * Tweak individual bids per the preferences.
   */
  if (bid < S.prefMinInd)
    bid = S.prefMinInd;
  if (bid > S.prefMaxInd)
    bid = S.prefMaxInd;
  if (S.prefPartner && bid > S.prefMaxTeam)
    bid = S.prefMaxTeam;

  MakeBid(p, bid);
}

/*
 * Determine what cards some player "p" may be holding in suit "s"
 * from the viewpoint of player "v".
 */
static int
SuitMap(int p, int v, int s)
{
  int map, i;

  if (p == v) {
    map = 0;
    for (i = 0; i < 13 && S.p[v].card[i] >= 0; i++)
      if (Suit(S.p[v].card[i]) == s)
	map |= (1 << Rank(S.p[v].card[i]));
  }
  else {
    map = S.p[p].suit[s];
    if (S.prefPartner && v == ((p+2) % 4))
      map &= ~(S.p[p].pmask[s]);
    for (i = 0; i < 13 && S.p[v].card[i] >= 0; i++)
      if (Suit(S.p[v].card[i]) == s)
	map &= ~(1 << Rank(S.p[v].card[i]));
  }
  return map;
}

static int
CardCmp(int c1, int c2)
{
  if (Suit(c1) == Suit(c2))
    return c1 - c2;
  else if (Suit(c1) == 0)
    return 1;
  else if (Suit(c2) == 0)
    return -1;
  else
    return 0;
}

int
ValidPlay(int p, int card)
{
  int i, suit, high, tmp;

  /*
   * Low clubs must be the first trick.
   */
  if (S.prefClubTrick && S.p[p].card[12] >= 0) {
    tmp = -1;
    for (i = 0; i < 13; i++)
      if (Suit(S.p[p].card[i]) == 2)
	if (tmp < 0 || S.p[p].card[i] < tmp)
	  tmp = S.p[p].card[i];
    if (tmp >= 0)
      return (card == tmp) ? 0 : invalidLowClub;
  }

  /*
   * Determine the suit which was led and the highest card played.
   */
  suit = high = -1;
  for (i = (p+1) % 4; i != p; i = (i+1) % 4)
    if (S.p[i].play >= 0) {
      if (suit < 0) {
	high = S.p[i].play;
	suit = Suit(high);
      }
      else if (CardCmp(S.p[i].play, high) > 0)
	high = S.p[i].play;
    }

  /*
   * Cannot lead trump until broken.
   */
  if (suit < 0) {
    if (S.prefSpadeBreak && Suit(card) == 0 && S.played[0] == 0) {
      for (i = 0; i < 13 && S.p[p].card[i] >= 0; i++)
	if (Suit(S.p[p].card[i]) != 0)
	  return invalidTrump;
    }
    return 0;
  }

  /*
   * Must follow suit.
   */
  if (Suit(card) != suit) {
    for (i = 0; i < 13 && S.p[p].card[i] >= 0; i++)
      if (Suit(S.p[p].card[i]) == suit)
	return invalidSuit;
  }

  /*
   * Must beat the highest card if possible.
   */
  if (S.prefMustBeat && CardCmp(card, high) <= 0) {
    tmp = 0;
    for (i = 0; i < 13 && S.p[p].card[i] >= 0; i++)
      if (Suit(S.p[p].card[i]) == suit) {
	tmp++;
	if (CardCmp(S.p[p].card[i], high) > 0)
	  return invalidMustBeat;
      }
    if (tmp == 0) {
      for (i = 0; i < 13 && S.p[p].card[i] >= 0; i++)
	if (Suit(S.p[p].card[i]) == 0 && CardCmp(S.p[p].card[i], high) > 0)
	  return invalidMustBeat;
    }
  }

  /*
   * Valid!
   */
  return 0;
}

int
ValidMisdeal(int p)
{
  int i, faces;
  int suitCount[4];

  if (!S.prefMisdeals)
    return 0;

  faces = 0;
  for (i = 0; i < 4; i++)
    suitCount[i] = 0;
  for (i = 0; i < 13; i++) {
    if (S.p[p].card[i] < 0)
      return 0;
    suitCount[Suit(S.p[p].card[i])]++;
    if (Rank(S.p[p].card[i]) >= 9)	// higher than jack
      faces++;
  }
  for (i = 0; i < 4; i++)
    if (suitCount[i] >= 7)
      return 1;
  if (suitCount[0] <= 1 || faces == 0)
    return 1;
  return 0;
}

int
CanBidBlindNil(int p)
{
  int i, myScore, hiScore;

  if (!S.prefBlindNils)
    return 0;
  if (!S.pref200Down)
    return 1;

  if (S.prefPartner) {
    myScore = S.p[p%2].score;
    hiScore = S.p[(p+1)%2].score;
  }
  else {
    hiScore = myScore = S.p[p].score;
    for (i = 0; i < 4; i++)
      if (S.p[i].score > hiScore)
	hiScore = S.p[i].score;
  }
  return (myScore + 200 > hiScore) ? 0 : 1;
}

void
Score(int p, struct score *pscore)
{
  int pard, bags;

  if (S.prefPartner) {
    pard = (p + 2) % 4;
    pscore->bid = S.p[p].bid + S.p[pard].bid;
    pscore->taken = 0;
    pscore->nilScore = 0;
    pscore->bags = 0;

    if (S.p[p].bid > 0)
      pscore->taken += S.p[p].taken;
    else if (S.p[p].taken == 0)
      pscore->nilScore += (S.p[p].blind) ? 200 : 100;
    else {
      pscore->nilScore -= 100;
      pscore->bags += S.p[p].taken;
    }

    if (S.p[pard].bid > 0)
      pscore->taken += S.p[pard].taken;
    else if (S.p[pard].taken == 0)
      pscore->nilScore += (S.p[pard].blind) ? 200 : 100;
    else {
      pscore->nilScore -= 100;
      pscore->bags += S.p[pard].taken;
    }

    if (pscore->bid > 0 && pscore->taken >= pscore->bid) {
      pscore->bidScore = 10 * (int)pscore->bid + (pscore->taken - pscore->bid);
      if (S.prefBonusBid > 0 && pscore->bid >= S.prefBonusBid)
	pscore->bidScore += 10 * S.prefBonusBid;
      pscore->bags += pscore->taken - pscore->bid;
    }
    else {
      pscore->bidScore = -10 * (int)pscore->bid;
    }
  }
  else {
    pscore->bid = S.p[p].bid;
    pscore->taken = S.p[p].taken;

    if (pscore->bid > 0) {
      pscore->nilScore = 0;
      if (pscore->taken >= pscore->bid) {
	pscore->bags = pscore->taken - pscore->bid;
	pscore->bidScore = 10 * (int)pscore->bid + pscore->bags;
	if (S.prefBonusBid > 0 && pscore->bid >= S.prefBonusBid)
	  pscore->bidScore += 10 * S.prefBonusBid;
      }
      else {
	pscore->bags = 0;
	pscore->bidScore = -10 * (int)pscore->bid;
      }
    }
    else if (pscore->taken == 0) {
      pscore->nilScore = (S.p[p].blind) ? 200 : 100;
      pscore->bags = 0;
      pscore->bidScore = 0;
    }
    else {
      pscore->nilScore = -100;
      pscore->bags = S.p[p].taken;
      pscore->bidScore = 0;
    }
  }

  pscore->bagScore = 0;
  for (bags = S.p[p].bags + pscore->bags; bags >= 10; bags -= 10)
    pscore->bagScore -= 100;
  pscore->totalScore = pscore->bidScore + pscore->nilScore + pscore->bagScore;
}

void
PlayCard(int p, int card)
{
  int i, s, r, lead, suit, high;
  struct score score;

#ifdef DEBUG
  printf("Player %d played %s\n", p, CardName(card));
#endif

  /*
   * Remove card from player's hand.
   */
  for (i = 0; i < 13; i++)
    if (S.p[p].card[i] == card)
      break;
  for (; i < 12; i++)
    S.p[p].card[i] = S.p[p].card[i+1];
  S.p[p].card[12] = -1;

  /*
   * Determine the suit which was led and the highest card played
   * so far.
   */
  suit = high = -1;
  for (i = (p+1) % 4; i != p; i = (i+1) % 4)
    if (S.p[i].play >= 0) {
      if (suit < 0) {
	high = i;
	suit = Suit(S.p[i].play);
      }
      else if (CardCmp(S.p[i].play, S.p[high].play) > 0)
	high = i;
    }

  /*
   * Set state variables.
   */
  s = Suit(card);
  r = Rank(card);
  S.p[p].play = card;
  S.played[s] |= (1<<r);
  if (suit >= 0 && s != suit) {
    S.p[p].suit[suit] = 0;	// he must be void
    if (S.prefExpert && S.p[p].bid == 0 && S.p[p].taken == 0 && s != 0) {
      // He should have sluffed his highest in this suit.
      for (i = r + 1; i < 13; i++)
	S.p[p].suit[s] &= ~(1<<i);
    }
  }

  if (S.prefExpert && s == suit && S.p[p].bid == 0 && S.p[p].taken == 0) {
    if (!(S.p[p].suit[s] & (1<<r))) {
      /*
       * If we didn't think this player could have this card, then he's
       * doing something funky and all bets are off.  He might have anything.
       */
      S.p[p].suit[s] = (~S.played[s]) & 0x1fff;
    }
    else if (high >= 0 && card < S.p[high].play) {
      /*
       * nil bids don't have any between this card and highest play
       */
      for (i = r + 1; i < Rank(S.p[high].play); i++)
	S.p[p].suit[s] &= ~(1<<i);
    }
  }
  if (S.p[p].pmask[s] & (1<<r)) {
    /*
     * He didn't trade us his highest or lowest?  Hmmm...
     */
    S.p[p].pmask[s] = 0;
  }

  /*
   * Ok, nobody has this card any more.
   */
  for (i = 0; i < 4; i++)
    S.p[i].suit[s] &= ~(1<<r);

  /*
   * If everyone has played, finish the trick.
   */
  lead = (p + 1) % 4;
  if (S.p[lead].play >= 0) {
    for (p = lead, i = (lead+1) % 4; i != lead; i = (i+1) % 4)
      if (CardCmp(S.p[i].play, S.p[p].play) > 0)
	p = i;
    S.p[p].taken++;

    /*
     * Update last trick info.
     */
    for (i = 0; i < 4; i++) {
      S.p[i].last = S.p[i].play;
      if (i == p)
	S.p[i].last |= lastWon;
    }

    /*
     * If this is not the last trick, the winner leads next.  If this
     * IS the last trick, sort out the score and indicate that it's
     * time to redeal.
     */
    if (S.p[0].card[0] >= 0) {
      S.turn = p;
      S.state = StEndTrick;
    }
    else {
      S.state = StEndHand;
    }
  }
  else {
    /*
     * Otherwise, update the turn indicator.
     */
    if (++(S.turn) >= 4)
      S.turn = 0;
    S.state = SetState();
  }
}

static int suit, high, pard, plays, agg, lastTrick;
static struct play {
  s8bit card;		// card's value
  s8bit trick;		// likelyhood card will take this trick
  s8bit future;		// likelyhood card will take future trick
} play[13];

static void
Calculate(int p, struct play *play)
{
  int i, mask, r, o, map, s, count, danger, trump, n, cover, sCount;

  /*
   * The scale for all likelihoods runs as follows:
   *
   * -1      Guaranteed not to take trick
   * 0..13   Unlikely to take trick (based on rank)
   * 50..63  Likely to take trick (based on rank)
   * 100     Guaranteed to take trick
   */
  for (mask = 0, r = Rank(play->card) + 1; r < 13; r++)
    mask |= (1<<r);		// bitmap of ranks which beat this card

  /*
   * If this is the highest trump then its future is just as bright
   * as its present.
   */
  if (Suit(play->card) == 0 && (S.played[0] & mask) == mask) {
    play->future = 100;
    if (high >= 0 && CardCmp(play->card, S.p[high].play) < 0)
      play->trick = -1;
    else
      play->trick = 100;
    return;
  }

  /*
   * Card's rank is its likelyhood of taking this trick if the card
   * is higher than the highest current card.
   */
  if (high < 0
  || CardCmp(play->card, S.p[high].play) > 0
  || play->card == S.p[high].play) {
    count = sCount = 0;
    o = (p+1) % 4;		// next player
    if (S.p[o].play >= 0) {	// we're the last card
      play->trick = 100;
    }
    else {
      if (suit < 0) {
	s = Suit(play->card);
	trump = 0;
      }
      else {
	s = suit;
	trump = (s != 0 && Suit(play->card) == 0);	// we trumped
      }

      for (; o != p && S.p[o].play < 0; o = (o+1) % 4) {
	if (o == pard && S.p[p].bid > 0)
	  continue;
	map = SuitMap(o, p, s);
	if (map == 0 && s != 0) {	// void in the lead suit
	  if ((map = SuitMap(o, p, 0)) != 0) {
	    trump = 0;	// our trump isn't guaranteed anymore
	    if (Suit(play->card) != 0 || (map & mask))
	      count++;
	  }
	}
	else if (s == Suit(play->card) && (map & mask)) {
	  count++;
	  sCount++;
	}
      }
      if (trump)
	play->trick = 50;
      else if (count) {
	play->trick = Rank(play->card);
	if (sCount == 0 && agg <= 25)
	  play->trick += 15;	// opponents may not trump
      }
      else
	play->trick = Rank(play->card) + 60;
    }
  }
  else
    play->trick = -1;

  /*
   * Special case for prefMustBeat -- if the highest is out there then this card
   * may not be good.
   */
  if (S.prefMustBeat) {
    s = Suit(play->card);
    for (r = Rank(play->card) + 1; r < 13 && play->trick >= 0; r++) {
      count = 0;
      for (o = (p+3) % 4; o != p && S.p[o].play >= 0; o = (o+3) % 4)
	if (SuitMap(o, p, s) & (1<<r)) {
	  count++;	// a player who's played may have this higher card
	  break;
	}
      if (count == 0) {
	for (o = (p+1) % 4; o != p && S.p[o].play < 0; o = (o+1) % 4)
	  if (SuitMap(o, p, s) & (1<<r)) {
	    play->trick = -1;	// they gotta beat it
	    break;
	  }
      }
    }
  }

  /*
   * Figure out how many players probably still have cards in
   * this suit, and determine likelihood of making card good in
   * future tricks.
   */
  s = Suit(play->card);
  count = danger = trump = cover = 0;
  for (o = (p+1) % 4; o != p; o = (o+1) % 4) {
    if ((map = SuitMap(o, p, s)) != 0)
      count++;
    if (o == pard && S.p[p].bid > 0)
      continue;
    if (map & mask)
      danger++;
    else if (map == 0 && s != 0 && SuitMap(o, p, 0) != 0)
      trump++;
  }
  if (trump)
    play->future = Rank(play->card);
  else {
    /*
     * Count the cards in this suit we know about.  Then average
     * the remaining cards per player.
     */
    n = 13;
    for (r = 0; r < 13; r++)
      if (S.played[s] & (1<<r))
	n--;
    for (r = 0; r < 13 && S.p[p].card[r] >= 0; r++)
      if (Suit(S.p[p].card[r]) == s) {
	cover++;	// XXX - does ace cover king?
	n--;
      }

    n = (count == 0) ? 0 : ((n * 100) / count);
    if (n <= 100)
      danger++;

    /*
     * XXX this should be redone to use relative rank in remaining
     * cards rather than hard-coded ace, king, queen.
     */
    play->future = Rank(play->card) + ((s == 0) ? 15 : 0);
    if (Rank(play->card) == 12) {		// ace
      if (n > 100)
	play->future += 50;
    }
    else if (Rank(play->card) == 11) {	// king
      if (!danger || (cover >= 1 && n > 200))
	play->future += 50;
      else if (cover >= 1 && n > 100)
	play->future += 30;
    }
    else if (Rank(play->card) == 10) {	// queen
      if (!danger || (cover >= 2 && n > 300))
	play->future += 30;
    }
  }
}

static int
PlayNil(int p)
{
  int i, chosen = -1;

  /*
   * For nil bids, pick the card with highest potential that doesn't
   * take the trick.
   */
  for (i = 0; i < plays; i++) {
    if (chosen < 0 || play[i].trick < play[chosen].trick)
      chosen = i;
    else if (play[i].trick == play[chosen].trick) {
      if (play[i].future > play[chosen].future)
	chosen = i;
    }
  }
  return chosen;
}

static int
CoverNil(int p)
{
  int i, chosen = -1, mask, r, danger = 0, sluff = 0, s;
  int map[4];

  /*
   * Construct pard's suitmaps.
   */
  for (i = 0; i < 4; i++)
    map[i] = SuitMap(pard, p, i);

  /*
   * If our pard has the highest card, try to beat it.
   */
  if (pard == high) {
    for (i = 0; i < plays; i++)
      if (CardCmp(play[i].card, S.p[pard].play) > 0)
	if (chosen < 0 || CardCmp(play[i].card, play[chosen].card) < 0)
	  chosen = i;
  }

  /*
   * Determine if there's any danger of our pard taking this trick.
   */
  if (high >= 0 && S.p[pard].play < 0 && Suit(S.p[high].play) == suit) {
    for (r = Rank(S.p[high].play) + 1; r < 13; r++)
      if (map[suit] & (1<<r))
	danger++;
  }

  /*
   * If there's no danger or we can't cover him, sluff intelligently.
   */
  if (chosen < 0 && danger == 0 && high >= 0) {
    if (agg <= 50)
      sluff = 1;
    // else we'll drop through and play normally
  }
  else if (chosen < 0) {
    /*
     * If we're here, then our pard hasn't played yet and there's some danger of
     * his taking the trick.  First things first -- always try to lead to a void.
     */
    if (high < 0) {
      for (i = 0; i < plays; i++)
	if (map[Suit(play[i].card)] == 0) {
	  if (chosen < 0)
	    chosen = i;
	  else if (agg < 30 && play[i].trick < play[chosen].trick)
	    chosen = i;
	  else if (agg >= 30 && play[i].future <= play[i].trick) {
	    if (play[chosen].future > play[chosen].trick)
	      chosen = i;
	    else if (play[i].trick > play[chosen].trick)
	      chosen = i;
	    else if (play[i].trick == play[chosen].trick) {
	      if (play[i].future < play[chosen].future)
		chosen = i;
	    }
	  }
	}
    }

    /*
     * Try to pick something he can't beat.
     */
    if (chosen < 0) {
      for (i = 0; i < plays; i++) {
	s = Suit(play[i].card);
	if (high >= 0 && CardCmp(play[i].card, S.p[high].play) <= 0)
	    continue;	// don't consider cards which don't beat high card

	if (high < 0 || s == suit) {
	  for (mask = 0, r = Rank(play[i].card) + 1; r < 13; r++)
	    mask |= (1<<r);		// bitmap of ranks which beat this card
	  if (map[s] & mask)
	    continue;
	}
	if (chosen < 0)
	  chosen = i;
	else if (agg < 30) {
	  if (s == suit) {
	    if (play[i].trick < play[chosen].trick)
	      chosen = i;
	  }
	  else {
	    // If we're here then we're trumping.  Normally we'd throw
	    // our highest trump at this low aggression level, but we
	    // need to cover our pard's trumps later.
	    //
	    // XXX maybe throw a higher trump if pard is covered?
	    if (play[i].trick < play[chosen].trick)
	      chosen = i;	// for now, choose lowest trump
	    else if (play[i].trick == play[chosen].trick)
	      if (play[i].future < play[chosen].future)
		chosen = i;
	  }
	}
	else if (agg >= 30) {
	  if (s != suit) {
	    if (play[i].trick < play[chosen].trick)
	      chosen = i;	// both must be trumps -- choose lower
	    else if (play[i].trick == play[chosen].trick)
	      if (play[i].future < play[chosen].future)
		chosen = i;
	  }
	  else if (play[i].future <= play[i].trick) {
	    if (play[chosen].future > play[chosen].trick)
	      chosen = i;
	    else if (play[i].trick > play[chosen].trick)
	      chosen = i;
	    else if (play[i].trick == play[chosen].trick) {
	      if (play[i].future < play[chosen].future)
		chosen = i;
	    }
	  }
	}
      }
    }

    /*
     * If we still haven't chosen one, then we're in a load of trouble.
     * Play our highest card.
     */
    if (chosen < 0 && sluff == 0) {
      sluff = 1;
      for (i = 0; i < plays; i++)
	if (suit < 0 || Suit(play[i].card) == suit)
	  if (chosen < 0 || Rank(play[i].card) > Rank(play[chosen].card))
	    chosen = i;
    }

    /*
     * If this card is essentially one card greater than the current
     * high, then just sluff.
     */
    if (chosen >= 0 && high >= 0
     && Suit(play[chosen].card) == Suit(S.p[high].play)) {
      for (r = Rank(S.p[high].play) + 1; S.played[s] & (1<<r); r++)
	continue;
      if (r == Rank(play[chosen].card)) {
	sluff = 1;
	chosen = -1;
      }
    }
  }

  if (chosen < 0 && sluff) {
    // XXX sluff something intelligently
    for (i = 0; i < plays; i++)
      if (chosen < 0 || play[i].future < play[chosen].future)
	chosen = i;
  }

// XXX can you upgrade the chosen card to one with a better chance because you
// know that your pard doesn't have anything between this card and the higher one?

  return chosen;
}

static int
SetNil(int p)
{
  int i, pp, chosen = -1, r, mask, s, map, count;

  /*
   * If one of our opponents bid nil and either hasn't played or has played
   * the high card, try to get under it.
   */
  for (pp = (p + 1) % 4; pp != p && chosen < 0; pp = (pp + 1) % 4) {
    if (pp == pard || S.p[pp].bid != 0 || S.p[pp].taken != 0)
      continue;
    if (high < 0 && agg <= 50) {
      /*
       * If we're leading, lead something small.
       */
      for (i = 0; i < plays; i++) {
	for (mask = 0, r = Rank(play[i].card) + 1; r < 13; r++)
	  mask |= (1<<r);		// bitmap of ranks which beat this card
	if ((SuitMap(pp, p, Suit(play[i].card)) & mask) == 0)
	  continue;			// he can't beat this one -- skip it
	if (chosen < 0 || Rank(play[i].card) < Rank(play[chosen].card))
	  chosen = i;
      }
    }
    else if (high == pp) {
      /*
       * Try to get under the high card.
       */
      for (i = 0; i < plays; i++)
	if (CardCmp(play[i].card, S.p[high].play) <= 0
	 && play[i].future < 40
	 && (chosen < 0 || (play[i].future > play[chosen].future)))
	  chosen = i;
    }
    else if (high >= 0 && S.p[pp].play < 0) {
      /*
       * Count how many cards we have under the high.
       */
      count = 0;
      for (i = 0; i < plays; i++)
	if (CardCmp(play[i].card, S.p[high].play) <= 0)
	  count++;

      /*
       * Try to get just over the high card under certain conditions:
       * 1. our aggression is high and our pard's not the high card
       * 2. we don't have at least two cards under the high card
       * 3. the high card is a face card
       * 4. the nil bidder can't beat the high card
       */
      s = Suit(S.p[high].play);
      r = Rank(S.p[high].play);
      map = SuitMap(pp, p, suit);
      for (mask = 0, i = r + 1; i < 13; i++)
	mask |= (1<<i);		// bitmap of ranks which beat high card

      if ((agg > 50 && high != pard)
       || (count < 2)
       || (r >= 9 && high != pard)
       || (s != suit || (map & mask) == 0)) {
	if (s != suit) {	// high player trumped
	  if (map == 0)
	    map = SuitMap(pp, p, 0);
	  else
	    map = 0;		// play our highest trump
	}
	for (r++; r < 13; r++) {
	  if (!(S.played[s] & (1<<r))) {
	    for (i = 0; i < plays; i++)
	      if (play[i].card == (s * 13 + r)) {
		if (agg < 50 || play[i].future < 40 || S.prefPartner)
		  chosen = i;
		break;
	      }
	  }
	  if (map & (1<<r))
	    break;
	}
	/*
	 * If we can't get above the high card in suit, maybe we can
	 * trump it.
	 */
	if (chosen < 0 && s == suit && (map & mask) == 0) {
	  for (i = 0; i < plays; i++)
	    if (Suit(play[i].card) == 0)
	      if (agg < 50 || play[i].future < 40 || S.prefPartner)
		if (map != 0 || Rank(play[i].card) <= 6)
		  if (chosen < 0 || play[i].card > play[chosen].card)
		    chosen = i;
	}
      }
      /*
       * Otherwise, try to get under the high card.
       */
      for (i = 0; i < plays; i++)
	if (CardCmp(play[i].card, S.p[high].play) <= 0
	 && play[i].future < 40
	 && (chosen < 0 || (play[i].future > play[chosen].future)))
	  chosen = i;
    }
  }

  return chosen;
}

static int
PlayNormal(int p)
{
  int i, chosen = -1, n, r, s, tmp;
  struct play pCard;
  int pmap[4], omap[4];

  /*
   * If our pard has played, calculate the chances that his card has.
   */
  if (pard >= 0 && high == pard) {
    pCard.card = S.p[pard].play;
    Calculate(p, &pCard);
  }
  else
    pCard.trick = pCard.future = -1;

  /*
   * For normal play, our aggression level determines the desire we
   * have to take this trick.
   */
  if (lastTrick && pCard.trick >= 100)
    return PlayNil(p);
  if (agg == 0) {
    if (pard > 0 && S.p[pard].play == high && pCard.trick > 30) {
      /*
       * Our pard's winning the trick -- beat him if we can.
       */
    }
    else
      return PlayNil(p);
  }

  /*
   * If our aggression is zero then we're here because our pard is pro'ly
   * gonna win this trick -- win it for him instead with our highest and
   * bestest card.
   *
   * For any other aggression, find the card that can win the trick but
   * has the lowest future value.
   */
  for (i = 0; i < plays; i++) {
    if (agg > 0 && (play[i].future > play[i].trick ||
     (play[i].future == play[i].trick && (play[i].trick < 30 || !lastTrick))))
      continue;
    if (chosen < 0 || play[i].trick > play[chosen].trick)
      chosen = i;
    else if (play[i].trick == play[chosen].trick) {
      if (agg == 0 || (lastTrick && play[i].trick >= 50)) {
	if (play[i].future > play[chosen].future)
	  chosen = i;
      }
      else if (play[i].future < play[chosen].future)
	chosen = i;
    }
  }

  /*
   * Watch the finesse.  Beat an opponent's best card if we can do so.
   */
  if (chosen < 0 && high >= 0 && high != pard) {
    for (i = 0; i < plays; i++) {
      if (Suit(play[i].card) == Suit(S.p[high].play)
       && play[i].card > S.p[high].play
       && (chosen < 0 || play[i].card < play[chosen].card))
	chosen = i;
    }
  }

  /*
   * Don't play over our pard's good lead under normal circumstances.
   */
  if (agg > 25 && !lastTrick && chosen >= 0 && pCard.trick >= 50)
    chosen = -1;

  /*
   * Randomly elect not to take this trick.  If we can't afford to lose
   * one then our aggression would be higher.
   */
  if (agg > 0 && agg < 75 && chosen >= 0 && (pard < 0 || high != pard)
   && (SysRandom(0) % 100) > agg)
    chosen = -1;

  /*
   * Consider leading to our pard's void.
   */
  if (S.prefExpert && high < 0 && pard >= 0
   && SuitMap(pard, p, 0) != 0
   && (chosen < 0 || (SysRandom(0) % 2) == 0)) {
    tmp = -1;
    for (s = 0; s < 4; s++) {
      pmap[s] = SuitMap(pard, p, s);
      omap[s] = SuitMap((p+3)%4, p, s);
    }
    for (i = 0; i < plays; i++) {
      s = Suit(play[i].card);
      if (pmap[s] == 0 && omap[s] != 0)
	if (tmp < 0 || Rank(play[i].card) < Rank(play[tmp].card))
	  tmp = i;
    }
    if (tmp >= 0)
      chosen = tmp;
  }

  /*
   * If our aggression is pretty high and we still haven't chosen a
   * card, then consider playing a precious trump.
   */
  if (chosen < 0 && agg >= 75 && pCard.trick < 50) {
    for (i = 0; i < plays; i++)
      if (play[i].trick >= 50)
	if (chosen < 0 || Rank(play[i].card) < Rank(play[chosen].card))
	  chosen = i;
  }

  /*
   * If we haven't found a card yet, then nothing we have can take this
   * trick.  For a zero aggression, sluff our card which has the best
   * future.  For any other aggression, sluff our lowest card.
   */
  if (chosen < 0) {
    for (i = 0; i < plays; i++)
      if (chosen < 0)
	chosen = i;
      else if (agg == 0) {
	if (play[i].future > play[chosen].future)
	  chosen = i;
      }
      else if (play[i].future < play[chosen].future)
	chosen = i;
  }

  /*
   * Now, for mid-level aggressions we may opt to play a higher
   * card if we're pro'ly gonna lose this trick anyway.  This gets rid
   * of those pesky middle cards which are bag-machines later in the hand.
   */
  if (agg > 0 && agg < 75) {
    s = Suit(play[chosen].card);
    for (;;) {
      /*
       * Find the next highest card in suit.
       */
      n = -1;
      for (i = 0; i < plays; i++)
	if (Suit(play[i].card) == s
	 && play[i].card > play[chosen].card
	 && (n < 0 || play[i].card < play[n].card))
	    n = i;
      if (n < 0)
	break;

      /*
       * Determine the highest rank equivalence of the chosen card,
       * based on which cards have been played.
       */
      r = Rank(play[chosen].card) + 1;
      while (r < 13 && (S.played[s] & (1<<r))
	&& (high < 0 || (s*13+r) != S.p[high].play))
	  r++;
      if (r != Rank(play[n].card)		// cards are equivalent
       && ((play[n].future >= 50)		// card has a good future
        || (play[n].future >= 20 && agg >= 50))) // card has a decent future
	  break;
      chosen = n;
    }
  }

  return chosen;
}

static void
Play(int p)
{
  int i, chosen = -1, c;
  int myNeed, oppNeed, totTricks;

  /*
   * Determine the suit which was led and the highest card played
   * so far.
   */
  suit = high = -1;
  for (i = (p+1) % 4; i != p; i = (i+1) % 4)
    if (S.p[i].play >= 0) {
      if (suit < 0) {
	high = i;
	suit = Suit(S.p[i].play);
      }
      else if (CardCmp(S.p[i].play, S.p[high].play) > 0)
	high = i;
    }
  pard = (S.prefPartner) ? (p + 2) % 4 : -1;

  /*
   * Determine our "aggression" level.  We're more aggressive if we're
   * trying to take tricks or set our opponents.  We're less aggressive
   * if we're trying to bag'em.  Agressiveness is on a scale from 0 (least
   * aggressive) to 100 (most aggressive).
   */
  totTricks = 13 - S.p[0].taken - S.p[1].taken - S.p[2].taken - S.p[3].taken;
  if (pard < 0) {
    if ((myNeed = (int)S.p[p].bid - (int)S.p[p].taken) < 0)
      myNeed = 0;
    oppNeed = 0;
    for (i = 0; i < 4; i++)
      if (i != p && S.p[i].bid > S.p[i].taken)
	oppNeed += S.p[i].bid - S.p[i].taken;
  }
  else {
    myNeed = S.p[p].bid + S.p[pard].bid;
    if (S.p[p].bid > 0)
      myNeed -= S.p[p].taken;
    if (S.p[pard].bid > 0)
      myNeed -= S.p[pard].taken;
    if (myNeed < 0)
      myNeed = 0;

    i = (p + 1) % 2;
    oppNeed = S.p[i].bid + S.p[i+2].bid;
    if (S.p[i].bid > 0)
      oppNeed -= S.p[i].taken;
    if (S.p[i+2].bid > 0)
      oppNeed -= S.p[i+2].taken;
    if (oppNeed < 0)
      oppNeed = 0;
  }

// XXX agg = 0 for oppNeed == 0 && myNeed == totTricks && this trick hopeless

  if (myNeed == 0 && oppNeed == 0)	// can't set'em; bag'em instead
    agg = 0;
  else if (totTricks < myNeed || totTricks < oppNeed) // not enough tricks left
    agg = 0;
  else if (totTricks - myNeed - oppNeed >= 3)	// lotsa bags
    agg = (myNeed > 0) ? 25 : 0;
  else if (totTricks - myNeed - oppNeed >= 2)	// just be cautious
    agg = 50;
  else if (totTricks - myNeed - oppNeed >= 1)	// need tricks
    agg = 75;
  else						// really need tricks
    agg = 100;

  /*
   * The lastTrick indicator is a special case when this is the last trick
   * we need to complete our bid, yet there are too many bags out there
   * to warrant taking the trick with the lowest possible card.
   */
  if (S.prefExpert && myNeed == 1
   && (agg < 50 || oppNeed == 0 || (oppNeed <= 3 && agg <= 50)))
    lastTrick = 1;
  else if (S.prefExpert && myNeed == 0 && oppNeed == totTricks)
    lastTrick = 1;
  else
    lastTrick = 0;

  /*
   * Populate the play structure with our valid plays.
   */
  plays = 0;
  for (i = 0; i < 13 && S.p[p].card[i] >= 0; i++)
    if (ValidPlay(p, S.p[p].card[i]) == 0) {
      play[plays].card = S.p[p].card[i];
      Calculate(p, &play[plays]);
      plays++;
    }

#if defined(DEBUG) && 0
  printf("Calculate %d:", p);
  for (i = 0; i < plays; i++)
    printf(" %s,%d,%d", CardName(play[i].card), play[i].trick, play[i].future);
  printf("\n");
#endif

  /*
   * Now determine our disposition for this trick.
   */
  if (chosen < 0 && S.p[p].bid == 0 && (S.p[p].taken == 0 || oppNeed <= 0 || agg < 75))
    chosen = PlayNil(p);
  if (chosen < 0 && pard >= 0 && S.p[pard].bid == 0 && (agg <= 50 || S.p[pard].taken == 0))
    chosen = CoverNil(p);
  if (chosen < 0 && agg < 100)
    chosen = SetNil(p);
  if (chosen < 0)
    chosen = PlayNormal(p);

  /*
   * Play the chosen card.
   */
  PlayCard(p, play[chosen].card);
}

void
GameUpdate()
{
  int p;

  /*
   * Clean up state information.
   */
  switch (S.state) {
    case StEndTrick:
      NewTrick();
      break;

    case StEndHand:
      NewHand();
      break;

    case StEndGame:
      // nothing to do
      return;
  }
  S.state = 0;

  /*
   * Process next game turn.
   */
  if ((p = S.turn) == 0) {
    S.state = SetState();
  }
  else if (S.p[p].bid < 0) {
    /*
     * Make a bid.
     */
    Bid(p);
  }
  else {
    /*
     * Make a play.
     */
    Play(p);
  }
}

//---------------------------------------------------------------------------
// Everything from this point on is for debugging the bots on a UNIX
// system.
//---------------------------------------------------------------------------

#ifdef DEBUG
static int
Prompt(char *str)
{
  char line[80];

  fprintf(stderr, "%s: ", str);
  if (fgets(line, sizeof(line), stdin) == NULL)
    exit(0);
  return atoi(line);
}

void
BidChanged(int p1, int p2)
{
  printf("P%d's bid changed to %d as a result of P%d's bid\n",
    p1, S.p[p1].bid, p2);
}

void
DeclareMisdeal(int p)
{
  printf("P%d declared a === MISDEAL! ===\n", p);
}

void
CardsExchanged(int p)
{
  printf("P%d exchanged %d cards with his pard\n", p, S.prefBnilExch);
}

static void
Show()
{
  int p, i;

  printf("%15s  0  1  2  3  4  5  6  7  8  9 10 11 12\n", "");
  for (p = 0; p < 4; p++) {
    printf("P%d: %02d/%02d [%s] ", p, S.p[p].bid, S.p[p].taken,
      (S.p[p].play >= 0) ? CardName(S.p[p].play) : "--");

    if (p >= 0)
      for (i = 0; i < 13 && S.p[p].card[i] >= 0; i++)
	printf(" %s", CardName(S.p[p].card[i]));
    printf("\n");
  }
}

int
SysRandom(n)
  int n;
{
  if (n != 0) {
    printf("Seeding random() with %d\n", n);
    srandom(n);
  }
  return random();
}

int
main(int argc, char **argv)
{
  int c, seed, p;
  struct score score;

  seed = (argc > 1) ? atoi(argv[1]) : getpid();
  SysRandom(seed);

  SetDefaultPrefs(-1);
  S.prefGame = 0;
  S.prefExpert = 1;
  NewGame();
  while (S.state != StEndGame) {
    switch (S.state) {
      case StEndTrick:
	// printf("------ new trick\n");
	NewTrick();
	Show();
	break;

      case StEndHand:
	printf("====== Score\n");
	c = (S.prefPartner) ? 2 : 4;
	for (p = 0; p < c; p++) {
	  printf("P%d: prev score %d bags %d\n", p, S.p[p].score, S.p[p].bags);
	  Score(p, &score);
	  if (score.bid > 0)
	    printf("  Bid %d Made %d: %d\n",
	      score.bid, score.taken, score.bidScore);
	  if (S.p[p].bid == 0 || (S.prefPartner && S.p[p+2].bid == 0))
	    printf("  Nil bids: %d\n", score.nilScore);
	  if (score.bags > 0) {
	    printf("  %d bags taken\n", score.bags);
	    if (score.bagScore)
	      printf("  10 bags accumulated: %d\n", score.bagScore);
	  }
	  printf("  New score %d bags %d\n",
	    S.p[p].score + score.totalScore,
	    (S.p[p].bags + score.bags) % 10);
	}
	printf("\n");

	NewHand();
	break;

      case StUserBid:
#ifdef AUTOPLAY
	Bid(0);
	if (S.p[S.turn].bid >= 0)
	  Show();
#else
	Show();
	c = Prompt("Enter bid");
	MakeBid(0, c);
#endif
	break;

      case StUserPlay:
#ifdef AUTOPLAY
	Play(0);
#else
	do {
	  c = Prompt("Enter card #");
	} while (ValidPlay(0, S.p[0].card[c]) != 0);
	PlayCard(0, S.p[0].card[c]);
#endif
	break;

      default:
	c = S.p[S.turn].bid;
	GameUpdate();
	if (c < 0 && S.p[S.turn].bid >= 0)
	  Show();
    }
  }
}
#endif
