// SpadeLover Spades -- A Spades Game
// Copyright (C) 1998 Britt Yenne
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// To contact the author, visit:  http://www.jump.net/~syenne/pilot/

#define SpadesAppID	'SPAD'
#define SpadesDBType	'Data'

typedef unsigned char	u8bit;
typedef signed char	s8bit;
typedef unsigned short	u16bit;
typedef signed short	s16bit;
typedef unsigned int	u32bit;
typedef signed int	s32bit;

#define Rank(id)	((id) % 13)
#define Suit(id)	((int)(id) / 13)

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

enum spadeState {
  StNone = 0,		// no state info
  StUserBid,		// user's turn to bid
  StBnilExch,		// user must exchange some cards with pard
  StUserPlay,		// user's turn to play
  StEndTrick,		// end of trick
  StEndHand,		// end of hand
  StEndGame		// game is over
};

struct state {
  // Player information
  //
  struct player {
    s16bit score;	// score
    u16bit suit[4];	// suit bitfields for potentially held cards
    u16bit pmask[4];	// suit bitmask for partner-specific knowledge
    u8bit blind : 1;	// set if bid is blind
    u8bit bags : 7;	// bags
    s8bit card[13];	// player's cards
    s8bit bid;		// current bid
    s8bit taken;	// current taken tricks
    s8bit play;		// card played on table
    s8bit last;		// last card played, maybe w/ lastWon bit
  } p[4];

  // Game information
  //
  s8bit dealer;		// player who is dealer (0 == user)
  s8bit turn;		// player whose turn it is
  s8bit state;		// state indicator -- enum spadeState above
  u16bit played[4];	// bitmaps of played cards in each suit

  // User preferences
  //
  u16bit prefRevSuits : 1;	// reverse suit ordering
  u16bit prefRevRanks : 1;	// reverse rank ordering
  u16bit prefGame : 3;		// game ID
  u16bit prefExpert : 1;	// expert level

  // Game preferences
  //
# define nGames 6
  struct _gpref {
    s16bit gprefWin;		// winning score
    s16bit gprefLose;		// losing score

    u32bit gprefMinInd : 3;	// minimum individual bid
    u32bit gprefMaxInd : 4;	// maximum individual bid
    u32bit gprefMinTeam : 3;	// minimum team bid
    u32bit gprefMaxTeam : 4;	// maximum team bid
    u32bit gprefPartner : 1;	// partner spades
    u32bit gprefMustBeat : 1;	// must beat highest card
    u32bit gprefMisdeals : 1;	// allow misdeals to be declared
# define maxExch 2
    u32bit gprefBnilExch : 2;	// number of cards to exchange on blind nils
    u32bit gprefSpadeBreak : 1;	// spades must be broken
    u32bit gprefClubTrick : 1;	// first trick is lowest clubs
    u32bit gprefBonusBid : 4;	// ind. or team bid which gets 2x bonus
    u32bit gprefOneNil : 1;	// one partner must bid nil
    u32bit gprefBidSpades : 1;	// must bid your spades count
    u32bit gpref200Down : 1;	// must be 200 down to bid blind nil
    u32bit gprefBlindNils : 1;	// blind nils allowed
  } gpref[nGames];

} S;

// Codes returned from ValidPlay()
//
enum invalidType {
  validPlay = 0,	// play is valid
  invalidTrump,		// can't lead unbroken trump
  invalidSuit,		// must follow suit
  invalidLowClub,	// first trick must be lowest clubs
  invalidMustBeat	// must beat the high card
};

// Preferences shortcuts
//
#define prefWin		gpref[S.prefGame].gprefWin
#define prefLose	gpref[S.prefGame].gprefLose
#define prefMinInd	gpref[S.prefGame].gprefMinInd
#define prefMaxInd	gpref[S.prefGame].gprefMaxInd
#define prefMinTeam	gpref[S.prefGame].gprefMinTeam
#define prefMaxTeam	gpref[S.prefGame].gprefMaxTeam
#define prefPartner	gpref[S.prefGame].gprefPartner
#define prefMustBeat	gpref[S.prefGame].gprefMustBeat
#define prefMisdeals	gpref[S.prefGame].gprefMisdeals
#define prefBnilExch	gpref[S.prefGame].gprefBnilExch
#define prefSpadeBreak	gpref[S.prefGame].gprefSpadeBreak
#define prefClubTrick	gpref[S.prefGame].gprefClubTrick
#define prefBonusBid	gpref[S.prefGame].gprefBonusBid
#define prefOneNil	gpref[S.prefGame].gprefOneNil
#define prefBidSpades	gpref[S.prefGame].gprefBidSpades
#define pref200Down	gpref[S.prefGame].gpref200Down
#define prefBlindNils	gpref[S.prefGame].gprefBlindNils

// Indicator for S.p[p].last that this person won the trick.
// Yeah, this should have been a bitfield; oh well.
//
#define lastWon		0x80

// This bid may be passed to MakeBid() to indicate a blind nil.
//
#define bidBlindNil	20

// Score structure as returned by Score()
//
struct score {
  u8bit bid;		// player's (or team's) bid
  u8bit taken;		// tricks taken
  s16bit bidScore;	// points earned or lost for non-nil bids
  s16bit nilScore;	// points earned or lost for nil bids

  u8bit bags;		// bags taken
  s8bit bagScore;	// points earned or lost for bagging

  s16bit totalScore;	// total points earned or lost for this hand
};

// I use UNIX (linux or AIX) to debug this thing.  Hey, who wouldn't.
//
#ifdef _AIX
# define DEBUG
#endif
#ifdef linux
# define DEBUG
#endif

// Prototypes
//
extern void	SetDefaultPrefs (int);			// game.c
extern void	NewGame		();			// game.c
extern void	NewHand		();			// game.c
extern void	NewTrick	();			// game.c
extern void	GameUpdate	();			// game.c
extern int	ValidPlay	(int, int);		// game.c
extern int	ValidMisdeal	(int);			// game.c
extern int	CanBidBlindNil	(int);			// game.c
extern void	MakeBid		(int, int);		// game.c
extern void	BnilExchange	(int, s8bit*);		// game.c
extern void	PlayCard	(int, int);		// game.c
extern void	Score		(int, struct score *);	// game.c
extern void	SortCards	(s8bit *, int);		// game.c

extern void	bzero		(void*,int);		// spades.c
extern void	BidChanged	(int,int);		// spades.c
extern void	DeclareMisdeal	(int);			// spades.c
extern void	CardsExchanged	(int);			// spades.c
