// SpadeLover Spades -- A Spades Game
// Copyright (C) 1998 Britt Yenne
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// To contact the author, visit:  http://www.jump.net/~syenne/pilot/

#pragma pack(2)
#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>

#include "header.h"
#include "spades.h"

#define cardID(n)	(1100+(n))

#define textY	38

static char *ranks = "23456789TJQKA";
static char *suits = "\220\217\216\215";

static int tableX[4], tableY[4];
static int firstStart = 0;
static int saveGame = -1;

static int exchCount = 0;
static int exchDone = 0;
static s8bit exchIndex[maxExch];

static char buf[80];
static char title[24];
static int titleGame = -1;

static void
SavePrefs()
{
  int old = S.prefGame;

  if (saveGame >= 0)
    S.prefGame = saveGame;
  PrefSetAppPreferences(SpadesAppID, 0, 1, &S, sizeof(S), TRUE);
  S.prefGame = old;
}

void
bzero(void *ptr, int count)
{
  int i;
  char *s = (char*)ptr;

  for (i = 0; i < count; i++)
    s[i] = 0;
}

static int
ReadPrefs()
{
  Word size;
  SWord rc;

  size = sizeof(S);
  rc = PrefGetAppPreferences(SpadesAppID, 0, &S, &size, TRUE);
  if (rc == noPreferenceFound || size != sizeof(S)) {
    bzero(&S, sizeof(S));
    SetDefaultPrefs(-1);
    NewGame();
    firstStart = 1;
  }
  SysRandom(TimGetTicks());	// seed random generator
  return 0;
}

static void
ReSort()
{
  int i;

  for (i = 0; i < 13 && S.p[0].card[i] >= 0; i++)
    continue;
  SortCards(S.p[0].card, i);
}

static int
StartApplication(void)
{
  int error;

  if ((error = ReadPrefs()) != 0)
    return error;

#define topHeight 20
#define formHeight (160 - topHeight - cardHeight)

  tableX[0] = 80 - (cardWidth / 2);
  tableY[0] = formHeight / 2 + topHeight + 2;
  tableX[1] = 80 - (cardWidth / 2) - 2 - cardWidth;
  tableY[1] = (formHeight - cardHeight) / 2 + topHeight;
  tableX[2] = tableX[0];
  tableY[2] = formHeight / 2 + topHeight - cardHeight - 2;
  tableX[3] = 80 + (cardWidth / 2) + 2;
  tableY[3] = tableY[1];

  ReSort();
  FrmGotoForm(formID_spades);
}

static char *
PlayerName(int p, int teamName)
{
  switch (p) {
    case 0:  return "You";
    case 1:  return (teamName ? "They" : "West");
    case 2:  return "North";
    default: return "East";
  }
}

static void
Clear(int x, int y, int w, int h)
{
  RectangleType rect;

  rect.topLeft.x = x;
  rect.topLeft.y = y;
  rect.extent.x = w;
  rect.extent.y = h;
  WinEraseRectangle(&rect, 0);
}

static int first = 0;

static void
DrawCard(FormPtr form, int id, int rank, int x, int y, int w, int h)
{
  int index;
  RectangleType rect;

  rect.topLeft.x = x + 1;
  rect.topLeft.y = y + 1;
  rect.extent.x = w - 2;
  rect.extent.y = h - 2;
  if (id >= 0) {
    index = FrmGetObjectIndex(form, cardID(id));
    FrmSetObjectPosition(form, index, rect.topLeft.x, rect.topLeft.y);
    FrmSetObjectBounds(form, index, &rect);
    FrmShowObject(form, index);
  }

  WinDrawRectangleFrame(rectangleFrame, &rect);
  if (S.p[0].bid == -1) {
    rect.topLeft.x += 2;
    rect.topLeft.y += 2;
    rect.extent.x -= 4;
    rect.extent.y -= 4;
    WinFillRectangle(&rect, 0);
  }
  else {
    FntSetFont(0);
    WinDrawChars(&ranks[Rank(rank)], 1, x+4, y+2);
    WinDrawChars(&suits[Suit(rank)], 1, x+2, y+13);
  }
}

static void
DrawCards(FormPtr form)
{
  int i, count, x, e, n;

  Clear(0, 160 - cardHeight, 160, cardHeight);
  for (count = 0; count < 13 && S.p[0].card[count] >= 0; count++)
    continue;
  count -= exchCount;
  for (i = count + exchCount; i < 13; i++)
    FrmHideObject(form, FrmGetObjectIndex(form, cardID(i)));

  if (count > 0) {
    if ((x = 80 - ((count - 1) * 12 + cardWidth) / 2) < 0)
      x = 0;
    n = 0;
    for (i = 0; i < 13 && S.p[0].card[i] >= 0; i++) {
      for (e = 0; e < exchCount; e++)
	if (i == exchIndex[e])
	  break;
      if (e >= exchCount) {
	DrawCard(form, i, S.p[0].card[i], x + n * 12, 160-cardHeight,
	  (n == count-1) ? cardWidth : 13, cardHeight);
	n++;
      }
    }
  }
}

static void
DrawBid(FormPtr form, int field, char *buf, char *name,
  s8bit *save, s8bit val, int blind)
{
  char *s = buf;
  FieldPtr ptr;

  if (first || *save != val) {
    while (*name)
      *(s++) = *(name++);
    *(s++) = ':';
    *(s++) = ' ';
    if (blind)
      *(s++) = 'b';
    if (val >= 0)
      StrIToA(s, (int)val);
    else
      *s = '\0';

    ptr = (FieldPtr)FrmGetObjectPtr(form, FrmGetObjectIndex(form, field));
    FldEraseField(ptr);
    FldSetTextPtr(ptr, buf);
    FldDrawField(ptr);
    *save = val;
  }
}

static void
DrawBids(FormPtr form)
{
  int i, p;
  static s8bit save[8];
  static char str[8][6];

  for (i = p = 0; p < 4; p++) {
    DrawBid(form, fieldID_P1bid + p, str[i], "B",
      &save[i], S.p[p].bid, S.p[p].blind), i++;
    DrawBid(form, fieldID_P1taken + p, str[i], "T",
      &save[i], S.p[p].taken, 0), i++;
  }
}

static int
ExchX(int i, int n)
{
  int width, sep = 2;

  width = n * (cardWidth + sep) - sep;
  return (160 - width) / 2 + (i * (cardWidth + sep));
}

static void
DrawTable(FormPtr form)
{
  int i;
  RectangleType rect;
  static s8bit save[4];
  static int oldExchCount = 0;

  /*
   * First handle exchanged cards.
   */
  if (first || oldExchCount != exchCount) {
    /*
     * Erase the old cards.
     */
    rect.topLeft.y = 65;
    rect.extent.x = cardWidth;
    rect.extent.y = cardHeight;
    for (i = 0; i < oldExchCount; i++) {
      rect.topLeft.x = ExchX(i, oldExchCount);
      WinEraseRectangle(&rect, 0);
    }

    /*
     * Draw the new.
     */
    for (i = 0; i < exchCount; i++) {
      DrawCard(form, exchIndex[i], S.p[0].card[exchIndex[i]],
	ExchX(i, exchCount), rect.topLeft.y,
	cardWidth, cardHeight);
    }
    oldExchCount = exchCount;
  }

  /*
   * And now for played cards.
   */
  if (S.state != StBnilExch && !exchDone) {
    for (i = 0; i < 4; i++)
      if (first || save[i] != S.p[i].play) {
	rect.topLeft.x = tableX[i];
	rect.topLeft.y = tableY[i];
	rect.extent.x = cardWidth;
	rect.extent.y = cardHeight;
	WinEraseRectangle(&rect, 0);

	if (S.p[i].play >= 0)
	  DrawCard(form, -1, S.p[i].play, tableX[i], tableY[i],
	    cardWidth, cardHeight);
	save[i] = S.p[i].play;
      }
  }
}

static void
DrawAll(FormPtr form)
{
  DrawBids(form);
  DrawCards(form);
  DrawTable(form);
}

static void
SelectForExchange(FormPtr form, int id)
{
  int e;

  for (e = 0; e < exchCount; e++)
    if (id == exchIndex[e])
      break;
  if (e < exchCount) {
    /*
     * Remove this card from exchange set.
     */
    exchCount--;
    for (; e < exchCount; e++)
      exchIndex[e] = exchIndex[e+1];
    FrmHideObject(form, FrmGetObjectIndex(form, buttonID_ok));
  }
  else if (exchCount < S.prefBnilExch) {
    exchIndex[exchCount++] = id;
    if (exchCount >= S.prefBnilExch)
      FrmShowObject(form, FrmGetObjectIndex(form, buttonID_ok));
  }
  else {
    //XXX display a bad boy message
    return;
  }

  DrawTable(form);
  DrawCards(form);
}

void
BidChanged(int p1, int p2)
{
  if (p1 == 0 && p2 == 0) {
    StrPrintF(buf,
      "Your bid changed to %d in accordance with the rules of this game.",
      S.p[0].bid);
  }
  else if (p1 == 0) {
    StrPrintF(buf,
      "Your bid changed to %d as a result of your partner's bid.",
      S.p[0].bid);
  }
  else if (p2 == 0) {
    StrPrintF(buf,
      "Your partner's bid changed to %d as a result of your bid.",
      S.p[p1].bid);
  }
  else if (p1 != p2) {
    StrPrintF(buf,
      "%s's bid changed to %d as a result of %s's bid.",
      PlayerName(p1, 0), S.p[p1].bid, PlayerName(p2, 0));
  }
  else
    return;	// internal bid change

  if (FrmCustomAlert(alertID_bidChanged, buf, NULL, NULL) == 1)
    FrmHelp(helpStr_bidChanged);
}

void
DeclareMisdeal(int p)
{
  int rc;
  char *wit;

  rc = FrmCustomAlert(alertID_misdeal, PlayerName(p, 0), NULL, NULL);
  if (rc > 0) {
    switch (SysRandom(0) % 4) {
      case 0:
	wit = "Well, it's only a game.";
	break;
      case 1:
	wit = "You can smack him.  Nobody's looking.";
	break;
      case 2:
	wit = "Your turn next time.";
	break;
      case 3:
	wit = "You can always set the slime.";
	break;
    }
    FrmCustomAlert(alertID_wit, wit, NULL, NULL);
  }
  if (FrmGetActiveFormID() == formID_spades)
    DrawAll(FrmGetActiveForm());
}

void
CardsExchanged(int p)
{
  int n;

  if (FrmGetActiveFormID() == formID_spades)
    DrawBids(FrmGetActiveForm());
  if (p != 0) {
    n = StrPrintF(buf, "%d card%s", S.prefBnilExch,
      ((S.prefBnilExch == 1) ? "" : "s"));
    buf[n] = '\0';	// just to be sure
    FrmCustomAlert(alertID_exchange, buf, NULL, NULL);
  }
}

static void
HandleBots(FormPtr form, int myFirst)
{
  int n;

  for (;;) {
    switch (S.state) {
      case StEndTrick:
      case StEndHand:
      case StEndGame:
	if (myFirst == 0) {
	  DrawTable(form);
	  SysTaskDelay((Long)SysTicksPerSecond());
	}
	if (S.state == StEndTrick) {
	  NewTrick();
	  DrawTable(form);
	}
	else {
	  FrmPopupForm(formID_score);
	  return;
	}
	break;

      case StUserBid:
	if (!CanBidBlindNil(0)) {
	  S.p[0].bid = -2;
	  DrawCards(form);
	}
	FrmPopupForm(formID_bid);
	return;

      case StBnilExch:
	Clear(30, textY, 100, 22);
	n = StrPrintF(buf, "Select %d card%s to trade", S.prefBnilExch,
	  ((S.prefBnilExch == 1) ? "" : "s"));
	WinDrawChars(buf, n, (160 - FntCharsWidth(buf, n)) / 2, textY);
	n = StrPrintF(buf, "with your partner");
	WinDrawChars(buf, n, (160 - FntCharsWidth(buf, n)) / 2, textY+11);
	return;

      case StUserPlay:
	FrmShowObject(form, FrmGetObjectIndex(form, labelID_play));
	return;

      default:
	/*
	 * Pause a bit -- bots gotta think, too.
	 */
	SysTaskDelay((Long)(SysTicksPerSecond() / 2));
	GameUpdate();
    }
    DrawBids(form);
    DrawTable(form);
    myFirst = 0;
  }
}

static Boolean
HandleMenu(int id, int *redraw, int *nextForm)
{
  int handled = 0, rc;

  *redraw = *nextForm = 0;
  switch (id)
  {
    case menuitemID_about:
      if (FrmAlert(alertID_about) == 1)
	FrmHelp(helpStr_license);
      handled = 1;
      break;

    case menuitemID_new:
      rc = FrmCustomAlert(alertID_confirm, "start a new game",
	NULL, NULL);
      if (rc == 0) {
	NewGame();
	*redraw = 1;
      }
      handled = 1;
      break;

    case menuitemID_score:
      *nextForm = formID_score;
      handled = 1;
      break;

    case menuitemID_prefs:
      *nextForm = formID_prefs;
      handled = 1;
      break;

    case menuitemID_last:
      *nextForm = formID_last;
      handled = 1;
      break;

    case menuitemID_rules:
      FrmHelp(helpStr_spades);
      handled = 1;
      break;
  }
  return handled;
}

static Boolean
spades(EventPtr event)
{
  FormPtr   form;
  int       handled = 0, rc, id, next, n, i;
  char      *msg, *msg2;
  s8bit     card[maxExch];

  switch (event->eType)
  {
    case frmOpenEvent:
    case frmUpdateEvent:
      form = FrmGetActiveForm();
      FrmHideObject(form, FrmGetObjectIndex(form, labelID_play));
      if (S.state != StBnilExch || exchDone) {
	if (exchDone)
	  ReSort();
	Clear(30, textY, 100, 22);
	exchCount = exchDone = 0;
      }
      else if (exchCount > S.prefBnilExch)
	exchCount = S.prefBnilExch;
      if (exchCount == 0 || exchCount < S.prefBnilExch)
	FrmHideObject(form, FrmGetObjectIndex(form, buttonID_ok));

      /*
       * Set the title to the game name.
       */
      if (S.prefGame != titleGame) {
	switch (S.prefGame) {
	  case 0:		msg = "Partner";	break;
	  case 1:		msg = "Solo";		break;
	  case 2:		msg = "Suicide";	break;
	  case 3:		msg = "Mirrors";	break;
	  case 4:		msg = "Custom 1";	break;
	  default:		msg = "Custom 2";
	}
	Clear(0, 0, 160, 15);
	StrPrintF(title, "Spades - %s", msg);
	FrmSetTitle(form, title);
	titleGame = S.prefGame;
      }

      /*
       * Draw the form.
       */
      FrmDrawForm(form);
      WinSetDrawWindow(FrmGetWindowHandle(form));

      buf[0] = buf[2] = buf[4] = buf[6] = 0xAA;  // 10101010
      buf[1] = buf[3] = buf[5] = buf[7] = 0x55;  // 01010101
      WinSetPattern((Word*)buf);

      first = 1;
      DrawAll(form);
      first = 0;

      if (firstStart) {
	FrmHelp(helpStr_start);
	firstStart = 0;
      }
      HandleBots(form, 1);

      handled = 1;
      break;

    case ctlSelectEvent:  // A control button was pressed and released.
      form = FrmGetActiveForm();
      id = event->data.ctlEnter.controlID;
      if (id >= cardID(0) && id <= cardID(12)
       && (S.state == StUserPlay || S.state == StBnilExch)) {
	id -= cardID(0);
	if (S.p[0].card[id] >= 0) {
	  if (S.state == StBnilExch) {
	    if (!exchDone) {
	      SelectForExchange(form, id);
	      handled = 1;
	    }
	  }
	  else if ((rc = ValidPlay(0, S.p[0].card[id])) == 0) {
	    PlayCard(0, S.p[0].card[id]);
	    FrmHideObject(form, FrmGetObjectIndex(form, labelID_play));
	    DrawAll(form);
	    HandleBots(form, 0);
	    handled = 1;
	  }
	  else {
	    switch (rc) {
	      case invalidTrump:
		msg = "You may not lead a spade until they've been broken";
		break;
	      case invalidSuit:
		msg = "You must follow suit";
		break;
	      case invalidLowClub:
		msg = "You must play your lowest club on the first trick";
		break;
	      case invalidMustBeat:
		msg = "You must beat the highest card played so far";
		break;
	      default:
		msg = NULL;
	    }
	    if (msg) {
	      if (rc == invalidSuit)
		msg2 = "This is a basic rule and is not configurable";
	      else
		msg2 = "If this bugs you, go to the preferences screen and "
		  "change the game options";
	      FrmCustomAlert(alertID_badPlay, msg, msg2, NULL);
	    }
	    handled = 1;
	  }
	}
      }
      else if (id == buttonID_ok) {
	if (!exchDone) {
	  for (i = 0; i < exchCount; i++)
	    card[i] = S.p[0].card[exchIndex[i]];
	  BnilExchange(0, card);

	  Clear(30, textY, 100, 22);
	  n = StrPrintF(buf, "Your pard passes back:");
	  WinDrawChars(buf, n, (160 - FntCharsWidth(buf, n)) / 2, textY+5);

	  exchDone = 1;
	  first = 1;
	  DrawTable(form);
	  first = 0;
	}
	else {
	  Clear(30, textY, 100, 22);
	  exchCount = exchDone = 0;
	  ReSort();
	  DrawTable(form);
	  DrawCards(form);
	  FrmHideObject(form, FrmGetObjectIndex(form, buttonID_ok));
	  HandleBots(form, 0);
	}
	handled = 1;
      }
      break;

    case menuEvent:
      handled = HandleMenu(event->data.menu.itemID, &rc, &next);
      if (rc > 0)
	FrmUpdateForm(formID_spades, 0);
      if (next)
	FrmPopupForm(next);
      break;
  }
  return handled;
}

static Boolean
bidding(EventPtr event)
{
  FormPtr   form;
  int       handled = 0, i, id, rc, next, pard, n;
  SWord     x, y;
  static int bid;

  switch (event->eType)
  {
    case frmOpenEvent:
      bid = -1;
      form = FrmGetActiveForm();
      if (S.prefMisdeals) {
	i = FrmGetObjectIndex(form, buttonID_ok);
	FrmGetObjectPosition(form, i, &x, &y);
	FrmSetObjectPosition(form, i, 10, y);
      }
      if (S.p[0].bid == -1) {
	FrmHideObject(form, FrmGetObjectIndex(form, buttonID_ok));
	FrmHideObject(form, FrmGetObjectIndex(form, buttonID_misdeal));
	for (i = 0; i <= 13; i++)
	  FrmHideObject(form, FrmGetObjectIndex(form, bidID_0 + i));
      }
      else {
	FrmHideObject(form, FrmGetObjectIndex(form, buttonID_dnil));
	FrmHideObject(form, FrmGetObjectIndex(form, buttonID_show));
	if (!S.prefMisdeals)
	  FrmHideObject(form, FrmGetObjectIndex(form, buttonID_misdeal));

	/*
	 * Determine the legal bids.
	 */
	if (S.prefBidSpades) {
	  for (n = i = 0; i < 13; i++)
	    if (Suit(S.p[0].card[i]) == 0)
	      n++;
	  for (i = 0; i <= 13; i++)
	    buf[i] = ((i == n) ? 1 : 0);
	}
	else {
	  for (i = 0; i <= 13; i++)
	    buf[i] = 1;		// keep track of legal bids
	  if (S.prefOneNil == 0 && S.prefMinInd > 0)
	    buf[0] = 0;
	  if (S.prefMaxInd >= S.prefMinInd
	   && (!S.prefPartner || S.prefMaxTeam >= S.prefMinTeam)) {
	    pard = (S.prefPartner && S.p[2].bid > 0) ? S.p[2].bid : -1;
	    for (i = 1; i <= 13; i++) {
	      if (((S.prefMinInd == 0 && i == 1)
	      || i == S.prefMinInd)
	      && pard + S.prefMinInd >= S.prefMaxTeam)
		continue;		// never FORCE us to bid nil
	      if (i < S.prefMinInd
	       || i > S.prefMaxInd
	       || (S.prefPartner
		&& ((S.prefOneNil && i < S.prefMinTeam)
		|| (i > S.prefMaxTeam)
		|| (!S.prefOneNil && pard >= 0 && (pard + i) > S.prefMaxTeam)
		|| (pard >= 0 && (pard + i) < S.prefMinTeam))))
		  buf[i] = 0;
	    }
	  }
	}

	/*
	 * Only display legal bids.
	 */
	id = 0;
	for (n = i = 0; i <= 13; i++) {
	  if (buf[i] == 0)
	    FrmHideObject(form, FrmGetObjectIndex(form, bidID_0 + i));
	  else {
	    n++;
	    id = i;
	  }
	}
	if (n <= 1) {
	  FrmHideObject(form, FrmGetObjectIndex(form, bidID_0 + id));
	  bid = id;
	}
      }
      FrmDrawForm(form);
      if (bid >= 0) {
	n = StrPrintF(buf, "Your bid is %d", bid);
	WinSetDrawWindow(FrmGetWindowHandle(form));
	WinDrawChars(buf, n, (115 - FntCharsWidth(buf, n)) / 2, 30);
      }
      handled = 1;
      break;

    case ctlSelectEvent:  // A control button was pressed and released.
      id = event->data.ctlEnter.controlID;
      if (id == buttonID_dnil) {
	MakeBid(0, bidBlindNil);
	FrmUpdateForm(formID_spades, 0);
	FrmReturnToForm(formID_spades);
	handled = 1;
      }
      else if (id == buttonID_show) {
	S.p[0].bid = -2;
	FrmUpdateForm(formID_spades, 0);
	FrmReturnToForm(formID_spades);
	handled = 1;
      }
      else if (id == buttonID_ok && bid >= 0) {
	MakeBid(0, bid);
	FrmUpdateForm(formID_spades, 0);
	FrmReturnToForm(formID_spades);
	handled = 1;
      }
      else if (id >= bidID_0 && id <= bidID_13) {
	bid = id - bidID_0;
	handled = 1;
      }
      else if (id == buttonID_misdeal) {
	if (ValidMisdeal(0)) {
	  NewHand();
	  FrmUpdateForm(formID_spades, 0);
	  FrmReturnToForm(formID_spades);
	}
	else
	  FrmAlert(alertID_badMisdeal);
	handled = 1;
      }
      break;

    case menuEvent:
      handled = HandleMenu(event->data.menu.itemID, &rc, &next);
      if (rc > 0) {
	FrmUpdateForm(formID_spades, 0);
	FrmReturnToForm(formID_spades);
      }
      else if (next) {
	form = FrmGetActiveForm();
	FrmEraseForm(form);
	FrmDeleteForm(form);
	FrmPopupForm(next);
      }
      break;
  }
  return handled;
}

static Boolean
lastform(EventPtr event)
{
  FormPtr   form;
  int       handled = 0, id, p, x, y, font;

  switch (event->eType)
  {
    case frmOpenEvent:
      form = FrmGetActiveForm();
      FrmDrawForm(form);
      WinSetDrawWindow(FrmGetWindowHandle(form));

      for (p = 0; p < 4; p++)
	if (S.p[p].last != -1) {
	  DrawCard(form, -1, (((u8bit)S.p[p].last) & 0x3f),
	    tableX[p] - LASTF_X, tableY[p] - LASTF_Y,
	    cardWidth, cardHeight);
	  if (S.p[p].last & lastWon) {
	    switch (p) {
	      case 0:
		x = tableX[p] + (cardWidth / 2) - 5;
		y = tableY[p] + cardHeight + 2;
		font = 5;
		buf[0] = 2;
		break;
	      case 1:
		x = tableX[p] - 10;
		y = tableY[p] + (cardHeight / 2) - 5;
		font = 4;
		buf[0] = 2;
		break;
	      case 2:
		x = tableX[p] + (cardWidth / 2) - 5;
		y = tableY[p] - 10;
		font = 5;
		buf[0] = 1;
		break;
	      case 3:
		x = tableX[p] + cardWidth + 2;
		y = tableY[p] + (cardHeight / 2) - 5;
		font = 4;
		buf[0] = 3;
		break;
	    }
	    font = FntSetFont(font);
	    WinDrawChars(buf, 1, x - LASTF_X, y - LASTF_Y);
	    FntSetFont(font);
	  }
	}

      handled = 1;
      break;

    case ctlSelectEvent:  // A control button was pressed and released.
      id = event->data.ctlEnter.controlID;
      if (id == buttonID_ok) {
	FrmUpdateForm(formID_spades, 0);
	FrmReturnToForm(formID_spades);
	handled = 1;
      }
      break;
  }
  return handled;
}

static Boolean
scoreform(EventPtr event)
{
  FormPtr   form;
  int       handled = 0, id, x, y, textHeight, p, n, i;
  u8bit     win[4], lose[4], nwin, nlose;
  char      *name;
  struct score score;

  switch (event->eType)
  {
    case frmOpenEvent:
      form = FrmGetActiveForm();
      FrmDrawForm(form);
      WinSetDrawWindow(FrmGetWindowHandle(form));
      FntSetFont(0);

      x = 2;
      y = 15;
      textHeight = 11;
      for (p = 0; p < (S.prefPartner ? 2 : 4); p++) {
	name = PlayerName(p, S.prefPartner);
	if (S.state == StEndHand) {
	  /*
	   * Show end-of-hand detail scoring.
	   */
	  Score(p, &score);
	  n = StrPrintF(buf, "%s bid %d made %d: %d points",
	    name, score.bid, score.taken,
	    ((S.prefPartner || S.p[p].bid > 0)
	      ? score.bidScore : score.nilScore));
	  WinDrawChars(buf, n, x, y);
	  y += textHeight;

	  if (S.prefPartner) {
	    if (S.p[p].bid == 0 || S.p[p+2].bid == 0) {
	      n = StrPrintF(buf, "Nil bids: %d points",
		score.nilScore);
	      WinDrawChars(buf, n, x + 10, y);
	      y += textHeight;
	    }
	    if (score.bags > 0) {
	      n = StrPrintF(buf, "Bags taken: %d", score.bags);
	      if (score.bagScore) {
		n += StrPrintF(buf + n, ", penalty %d",
		  score.bagScore);
	      }
	      WinDrawChars(buf, n, x + 10, y);
	      y += textHeight;
	    }
	  }
	  n = StrPrintF(buf, "New score %d, bags %d",
	    S.p[p].score + score.totalScore,
	    (S.p[p].bags + score.bags) % 10);
	  WinDrawChars(buf, n, x + 10, y);
	  y += textHeight + (S.prefPartner ? 5 : 0);
	}
	else {
	  /*
	   * Show just the score.
	   */
	  n = StrPrintF(buf, "%s: score %d, bags %d",
	    name, S.p[p].score, S.p[p].bags);
	  WinDrawChars(buf, n, x + 10, y);
	  y += textHeight + 5;
	}
      }
      handled = 1;
      break;

    case ctlSelectEvent:  // A control button was pressed and released.
      id = event->data.ctlEnter.controlID;
      if (id == buttonID_ok) {
	if (S.state == StEndHand)
	  NewHand();
	if (S.state == StEndGame) {
	  nwin = nlose = 0;
	  for (p = 0; p < (S.prefPartner ? 2 : 4); p++) {
	    if ((S.prefLose > S.prefWin && S.p[p].score >= S.prefLose)
	     || (S.prefLose != 0 && S.p[p].score <= S.prefLose)) {
	      lose[nlose++] = p;
	    }
	    else if (S.p[p].score >= S.prefWin) {
	      if (nwin == 0 || S.p[p].score > S.p[win[0]].score) {
		win[0] = p;
		nwin = 1;
	      }
	      else if (nwin > 0 && S.p[p].score == S.p[win[0]].score)
		win[nwin++] = p;
	    }
	  }
	  if (nwin) {
	    n = StrPrintF(buf, "%s ", PlayerName(win[0], S.prefPartner));
	    for (i = 1; i < nwin; i++)
	      n += StrPrintF(buf+n, "and %s ", PlayerName(win[i], S.prefPartner));
	    if (nwin > 1)
	      name = "tie";
	    else if (S.prefPartner || win[0] == 0)
	      name = "win";
	    else
	      name = "wins";
	    n += StrPrintF(buf+n, "%s the game!", name);
	  }
	  else if (nlose) {
	    n = StrPrintF(buf, "%s ", PlayerName(lose[0], S.prefPartner));
	    for (i = 1; i < nlose; i++)
	      n += StrPrintF(buf+n, "and %s ", PlayerName(lose[i], S.prefPartner));
	    if (nlose > 1 || S.prefPartner || lose[0] == 0)
	      name = "lose";
	    else
	      name = "loses";
	    n += StrPrintF(buf+n, "%s the game.", name);
	  }
	  else
	    n = StrPrintF(buf, "Nobody wins.", name);
	  FrmCustomAlert(alertID_gameOver, buf, NULL, NULL);
	  NewGame();
	}
	FrmUpdateForm(formID_spades, 0);
	FrmReturnToForm(formID_spades);
	handled = 1;
      }
      break;
  }
  return handled;
}

static Boolean
prefsform(EventPtr event)
{
  FormPtr   form;
  int       handled = 0, id, rc, i;
  static int updated;

  switch (event->eType)
  {
    case frmOpenEvent:
      saveGame = S.prefGame;
      updated = 1;
    case frmUpdateEvent:
      form = FrmGetActiveForm();
      if (updated) {
	FrmSetControlGroupSelection(form, groupID_game,
	  gameID_partner + S.prefGame);
	FrmSetControlGroupSelection(form, groupID_level,
	  (S.prefExpert ? levelID_expert : levelID_normal));
	FrmSetControlGroupSelection(form, groupID_sortSuits,
	  (S.prefRevSuits ? sortID_spadesLast : sortID_spadesFirst));
	FrmSetControlGroupSelection(form, groupID_sortRanks,
	  (S.prefRevRanks ? sortID_acesLast : sortID_acesFirst));
	updated = 0;
      }
      FrmDrawForm(form);
      handled = 1;
      break;

    case ctlSelectEvent:  // A control button was pressed and released.
      id = event->data.ctlEnter.controlID;
      if (id == buttonID_ok) {
	form = FrmGetActiveForm();
	S.prefGame = FrmGetObjectId(form, FrmGetControlGroupSelection(form,
	  groupID_game)) - gameID_partner;
	if (S.prefGame != saveGame)
	  NewGame();
	saveGame = -1;
	S.prefExpert =
	  (FrmGetObjectId(form, FrmGetControlGroupSelection(form,
	  groupID_level)) == levelID_expert) ? 1 : 0;
	S.prefRevSuits =
	  (FrmGetObjectId(form, FrmGetControlGroupSelection(form,
	  groupID_sortSuits)) == sortID_spadesLast) ? 1 : 0;
	S.prefRevRanks =
	  (FrmGetObjectId(form, FrmGetControlGroupSelection(form,
	  groupID_sortRanks)) == sortID_acesLast) ? 1 : 0;

	ReSort();
	FrmUpdateForm(formID_spades, 0);
	FrmReturnToForm(formID_spades);
	handled = 1;
      }
      else if (id == buttonID_cancel) {
	S.prefGame = saveGame;
	saveGame = -1;
	FrmUpdateForm(formID_spades, 0);
	FrmReturnToForm(formID_spades);
	handled = 1;
      }
      else if (id == buttonID_restore) {
	rc = FrmCustomAlert(alertID_confirm,
	  "restore all default game and user preferences",
	  NULL, NULL);
	if (rc == 0) {
	  bzero(&S, sizeof(S));
	  SetDefaultPrefs(-1);
	  saveGame = S.prefGame;
	  NewGame();
	  updated = 1;
	  FrmUpdateForm(formID_prefs, 0);
	}
	handled = 1;
      }
      else if (id == buttonID_gameOptions) {
	form = FrmGetActiveForm();
	S.prefGame = FrmGetObjectId(form,
	  FrmGetControlGroupSelection(form, groupID_game))
	  - gameID_partner;
	if ((i = formID_partnerPrefs + S.prefGame) > formID_customPrefs)
	  i = formID_customPrefs;
	FrmPopupForm(i);
	handled = 1;
      }
      else if (id == helpID_partner) {
	FrmHelp(helpStr_partner);
	handled = 1;
      }
      else if (id == helpID_solo) {
	FrmHelp(helpStr_solo);
	handled = 1;
      }
      else if (id == helpID_suicide) {
	FrmHelp(helpStr_suicide);
	handled = 1;
      }
      else if (id == helpID_mirrors) {
	FrmHelp(helpStr_mirrors);
	handled = 1;
      }
      else if (id == helpID_custom1 || id == helpID_custom2) {
	FrmHelp(helpStr_custom);
	handled = 1;
      }
      break;
  }
  return handled;
}

static int
SafeGetObjectIndex(FormPtr form, Word id)
{
  int i, count;

  count = FrmGetNumberOfObjects(form);
  for (i = 0; i < count; i++)
    if (FrmGetObjectId(form, i) == id)
      return i;
  return -1;
}

static void
FreePrefField(FormPtr form, Word id)
{
  int index;
  FieldPtr fld;
  VoidHand handle;

  if ((index = SafeGetObjectIndex(form, id)) >= 0) {
    fld = (FieldPtr)FrmGetObjectPtr(form, index);
    if ((handle = (VoidHand)FldGetTextHandle(fld)) != 0) {
      FldSetTextHandle(fld, (Handle)0);
      MemHandleFree(handle);
    }
  }
}

static void
SetPrefField(FormPtr form, Word id, int val)
{
  int index, max;
  FieldPtr fld;
  VoidHand handle;

  if ((index = SafeGetObjectIndex(form, id)) >= 0) {
    fld = (FieldPtr)FrmGetObjectPtr(form, index);
    if ((handle = (VoidHand)FldGetTextHandle(fld)) != 0)
      max = MemHandleSize(handle) - 1;
    else {
      if ((max = FldGetMaxChars(fld)) <= 0)
	max = 10;
      handle = MemHandleNew(max+1);
    }
    StrIToA((char*)MemHandleLock(handle), val);
    MemHandleUnlock(handle);
    FldSetText(fld, handle, 0, max);
  }
}

static void
SetPrefList(FormPtr form, Word listId, Word triggerId, int item)
{
  int index;
  ListPtr list;
  ControlPtr ctl;

  if ((index = SafeGetObjectIndex(form, listId)) >= 0) {
    list = (ListPtr)FrmGetObjectPtr(form, index);
    LstSetSelection(list, item);
    if ((index = SafeGetObjectIndex(form, triggerId)) >= 0) {
      ctl = (ControlPtr)FrmGetObjectPtr(form, index);
      CtlSetLabel(ctl, LstGetSelectionText(list, item));
    }
  }
}

static void
SetPrefToggle(FormPtr form, Word id, Boolean set)
{
  int index;

  if ((index = SafeGetObjectIndex(form, id)) >= 0)
    FrmSetControlValue(form, index, set);
}

static int
GetPrefField(FormPtr form, Word id, int val)
{
  int index;
  FieldPtr fld;
  VoidHand handle;

  if ((index = SafeGetObjectIndex(form, id)) >= 0) {
    fld = (FieldPtr)FrmGetObjectPtr(form, index);
    if ((handle = (VoidHand)FldGetTextHandle(fld)) != 0) {
      val = StrAToI((char*)MemHandleLock(handle));
      MemHandleUnlock(handle);
    }
  }
  return val;
}

static int
GetPrefList(FormPtr form, Word listId, int item)
{
  int index;
  ListPtr list;

  if ((index = SafeGetObjectIndex(form, listId)) >= 0) {
    list = (ListPtr)FrmGetObjectPtr(form, index);
    item = LstGetSelection(list);
  }
  return item;
}

static Boolean
GetPrefToggle(FormPtr form, Word id, Boolean set)
{
  int index;

  if ((index = SafeGetObjectIndex(form, id)) >= 0)
    set = FrmGetControlValue(form, index);
  return set;
}

static void
UpdateGamePrefs(FormPtr form, int gid)
{
  S.gpref[gid].gprefWin =
    GetPrefField(form, prefID_Win, S.gpref[gid].gprefWin);
  S.gpref[gid].gprefLose =
    GetPrefField(form, prefID_Lose, S.gpref[gid].gprefLose);

  S.gpref[gid].gprefMinInd =
    GetPrefList(form, prefID_MinInd, S.gpref[gid].gprefMinInd);
  S.gpref[gid].gprefMaxInd =
    GetPrefList(form, prefID_MaxInd, S.gpref[gid].gprefMaxInd);
  S.gpref[gid].gprefMinTeam =
    GetPrefList(form, prefID_MinTeam, S.gpref[gid].gprefMinTeam);
  S.gpref[gid].gprefMaxTeam =
    GetPrefList(form, prefID_MaxTeam, S.gpref[gid].gprefMaxTeam);
  S.gpref[gid].gprefBonusBid =
    GetPrefList(form, prefID_BonusBid, S.gpref[gid].gprefBonusBid);
  S.gpref[gid].gprefBnilExch =
    GetPrefList(form, prefID_BnilExch, S.gpref[gid].gprefBnilExch);

  S.gpref[gid].gprefPartner =
    GetPrefToggle(form, prefID_Partner, S.gpref[gid].gprefPartner);
  S.gpref[gid].gprefMustBeat =
    GetPrefToggle(form, prefID_MustBeat, S.gpref[gid].gprefMustBeat);
  S.gpref[gid].gprefMisdeals =
    GetPrefToggle(form, prefID_Misdeals, S.gpref[gid].gprefMisdeals);
  S.gpref[gid].gprefSpadeBreak =
    GetPrefToggle(form, prefID_SpadeBreak, S.gpref[gid].gprefSpadeBreak);
  S.gpref[gid].gprefClubTrick =
    GetPrefToggle(form, prefID_ClubTrick, S.gpref[gid].gprefClubTrick);
  S.gpref[gid].gprefOneNil =
    GetPrefToggle(form, prefID_OneNil, S.gpref[gid].gprefOneNil);
  S.gpref[gid].gprefBidSpades =
    GetPrefToggle(form, prefID_BidSpades, S.gpref[gid].gprefBidSpades);
  S.gpref[gid].gpref200Down =
    GetPrefToggle(form, prefID_200Down, S.gpref[gid].gpref200Down);
  S.gpref[gid].gprefBlindNils =
    GetPrefToggle(form, prefID_BlindNils, S.gpref[gid].gprefBlindNils);
  if (SafeGetObjectIndex(form, prefID_10for200) >= 0)
    S.gpref[gid].gprefBonusBid =
      (GetPrefToggle(form, prefID_10for200, 0)) ? 10 : 0;
  if (SafeGetObjectIndex(form, prefID_7for140) >= 0)
    S.gpref[gid].gprefBonusBid =
      (GetPrefToggle(form, prefID_7for140, 0)) ? 7 : 0;
}

static Boolean
gameprefsform(EventPtr event)
{
  FormPtr   form;
  int       handled = 0, id, rc, i;

  switch (event->eType)
  {
    case frmOpenEvent:
    case frmUpdateEvent:
      form = FrmGetActiveForm();

      SetPrefField(form, prefID_Win, S.prefWin);
      SetPrefField(form, prefID_Lose, S.prefLose);

      SetPrefList(form, prefID_MinInd, triggerID_MinInd, S.prefMinInd);
      SetPrefList(form, prefID_MaxInd, triggerID_MaxInd, S.prefMaxInd);
      SetPrefList(form, prefID_MinTeam, triggerID_MinTeam, S.prefMinTeam);
      SetPrefList(form, prefID_MaxTeam, triggerID_MaxTeam, S.prefMaxTeam);
      SetPrefList(form, prefID_BonusBid, triggerID_BonusBid, S.prefBonusBid);
      SetPrefList(form, prefID_BnilExch, triggerID_BnilExch, S.prefBnilExch);

      SetPrefToggle(form, prefID_Partner, S.prefPartner);
      SetPrefToggle(form, prefID_MustBeat, S.prefMustBeat);
      SetPrefToggle(form, prefID_Misdeals, S.prefMisdeals);
      SetPrefToggle(form, prefID_SpadeBreak, S.prefSpadeBreak);
      SetPrefToggle(form, prefID_ClubTrick, S.prefClubTrick);
      SetPrefToggle(form, prefID_OneNil, S.prefOneNil);
      SetPrefToggle(form, prefID_BidSpades, S.prefBidSpades);
      SetPrefToggle(form, prefID_200Down, S.pref200Down);
      SetPrefToggle(form, prefID_BlindNils, S.prefBlindNils);
      SetPrefToggle(form, prefID_10for200, (S.prefBonusBid == 10));
      SetPrefToggle(form, prefID_7for140, (S.prefBonusBid == 7));

      FrmDrawForm(form);
      handled = 1;
      break;

    case ctlSelectEvent:  // A control button was pressed and released.
      id = event->data.ctlEnter.controlID;
      if (id == buttonID_ok || id == buttonID_cancel) {
	form = FrmGetActiveForm();
	if (id == buttonID_ok)
	  UpdateGamePrefs(form, S.prefGame);

	FreePrefField(form, prefID_Win);
	FreePrefField(form, prefID_Lose);

	FrmUpdateForm(formID_prefs, 0);
	FrmReturnToForm(formID_prefs);
	handled = 1;
      }
      else if (id == buttonID_restore) {
	rc = FrmCustomAlert(alertID_confirm,
	  "restore default game preferences",
	  NULL, NULL);
	if (rc == 0) {
	  SetDefaultPrefs(S.prefGame);
	  FrmUpdateForm(FrmGetActiveFormID(), 0);
	}
	handled = 1;
      }
      else if (id == buttonID_copy1) {
	rc = FrmCustomAlert(alertID_confirm,
	  "copy this game's parameters to custom game 1",
	  NULL, NULL);
	if (rc == 0) {
	  form = FrmGetActiveForm();
	  S.gpref[nGames-2] = S.gpref[S.prefGame];
	  UpdateGamePrefs(form, nGames-2);
	}
	handled = 1;
      }
      else if (id == buttonID_copy2) {
	rc = FrmCustomAlert(alertID_confirm,
	  "copy this game's parameters to custom game 2",
	  NULL, NULL);
	if (rc == 0) {
	  form = FrmGetActiveForm();
	  S.gpref[nGames-1] = S.gpref[S.prefGame];
	  UpdateGamePrefs(form, nGames-1);
	}
	handled = 1;
      }
      else if (id >= helpID_Win && id <= helpStr_7for140) {
	FrmHelp(helpStr_Win + (id - helpID_Win));
	handled = 1;
      }
      break;
  }
  return handled;
}

static void
EventLoop(void)
{
  short err;
  int formID, tapDelay = 0;
  FormPtr form;
  EventType event;

  do {
    EvtGetEvent(&event, (tapDelay ? 5 : evtWaitForever));

    if (tapDelay && event.eType == nilEvent) {
      tapDelay = 0;
      continue;
    }

    if (SysHandleEvent(&event))
      continue;
    if (MenuHandleEvent(NULL, &event, &err))
      continue;

    if (event.eType == frmLoadEvent) {
      formID = event.data.frmLoad.formID;
      form = FrmInitForm(formID);
      FrmSetActiveForm(form);
      switch (formID) {
	case formID_spades:
	  FrmSetEventHandler(form, (FormEventHandlerPtr) spades);
	  break;
	case formID_bid:
	  FrmSetEventHandler(form, (FormEventHandlerPtr) bidding);
	  break;
	case formID_score:
	  FrmSetEventHandler(form, (FormEventHandlerPtr) scoreform);
	  break;
	case formID_last:
	  FrmSetEventHandler(form, (FormEventHandlerPtr) lastform);
	  break;
	case formID_prefs:
	  FrmSetEventHandler(form, (FormEventHandlerPtr) prefsform);
	  break;
	case formID_partnerPrefs:
	case formID_soloPrefs:
	case formID_suicidePrefs:
	case formID_mirrorsPrefs:
	case formID_customPrefs:
	  FrmSetEventHandler(form, (FormEventHandlerPtr) gameprefsform);
	  break;
      }
      continue;
    }

    if (event.eType == ctlSelectEvent) {
      if (tapDelay) {
	/*
	 * The user tapped two buttons just real fast.  This is
	 * almost certainly an accidental double-tap.
	 */
	continue;
      }
      tapDelay = 1;
    }

    /*
     * Dispatch the event to the active form event processing function.
     */
    FrmDispatchEvent(&event);

  } while (event.eType != appStopEvent);
}

static void
StopApplication(void)
{
  SavePrefs();
}

DWord
PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
  int error;

  if (cmd == sysAppLaunchCmdNormalLaunch) {
    if ((error = StartApplication()) != 0)
      return error;

    EventLoop();	// Event loop

    StopApplication();	// Application stop code
  }
  return 0;
}
